'''
A generic target tracking task
'''
from __future__ import barry_as_FLUFL
from multiprocessing.connection import wait
import numpy as np
import time
import os
import math
import traceback
import random
from collections import OrderedDict

from riglib.experiment import traits, Sequence, FSMTable, StateTransitions
from riglib.stereo_opengl import ik
from riglib import plants

from riglib.stereo_opengl.window import Window
from robot import trajectory
from .target_graphics import *

import matplotlib.pyplot as plt

## Plants
# List of possible "plants" that a subject could control either during manual or brain control
cursor = plants.CursorPlant()
shoulder_anchor = np.array([2., 0., -15])
chain_kwargs = dict(link_radii=.6, joint_radii=0.6, joint_colors=(181/256., 116/256., 96/256., 1), link_colors=(181/256., 116/256., 96/256., 1))
chain_20_20_endpt = plants.EndptControlled2LArm(link_lengths=[20, 20], base_loc=shoulder_anchor, **chain_kwargs)
chain_20_20 = plants.RobotArmGen2D(link_lengths=[20, 20], base_loc=shoulder_anchor, **chain_kwargs)

plantlist = dict(
    cursor=cursor,
    chain_20_20=chain_20_20,
    chain_20_20_endpt=chain_20_20_endpt)

class TargetTracking(Sequence):
    '''
    This is a generic cued target tracking skeleton, to form as a common ancestor to the most
    common type of motor control tracking tasks.
    '''
    status = dict(
        wait = dict(start_trial="trajectory", start_pause="pause"),
        wait_retry = dict(start_trial="trajectory", start_pause="pause"),
        trajectory = dict(enter_target="hold", timeout="timeout_penalty", start_pause="pause"),
        hold = dict(hold_complete="tracking_in", leave_target="hold_penalty", start_pause="pause"),
        tracking_in = dict(trial_complete="reward", leave_target="tracking_out", start_pause="pause"),
        tracking_out = dict(trial_complete="reward", enter_target="tracking_in", tracking_out_timeout="tracking_out_penalty", start_pause="pause"),
        timeout_penalty = dict(timeout_penalty_end="wait", start_pause="pause", end_state=True),
        hold_penalty = dict(hold_penalty_end="wait", hold_penalty_end_retry="wait_retry", start_pause="pause", end_state=True),
        tracking_out_penalty = dict(tracking_out_penalty_end="wait", start_pause="pause", end_state=True),
        reward = dict(reward_end="wait", start_pause="pause", stoppable=False, end_state=True),
        pause = dict(end_pause="wait", end_state=True)
        # all end_states will result in trial counter +1, so if you start pause during a penalty state, 
        # the next trial after unpausing will be current trial +2

    )

    # initial state
    state = "wait"
    tries = 0 # Helper variable to keep track of the number of failed attempts at initiating a given trial

    wait_time = traits.Float(1., desc="Length of time in wait state (inter-trial interval)")
    reward_time = traits.Float(.5, desc="Length of reward dispensation")
    timeout_time = traits.Float(10, desc="Time allowed to go between trajectories")
    timeout_penalty_time = traits.Float(1, desc="Length of penalty time for initiation timeout error")
    hold_time = traits.Float(.1, desc="Time of hold required at target before trajectory begins")
    hold_penalty_time = traits.Float(1, desc="Length of penalty time for target hold error")
    tracking_out_time = traits.Float(2.5, desc="Time allowed to be tracking outside the target") # AKA tolerance time
    tracking_out_penalty_time = traits.Float(1, desc="Length of penalty time for tracking out error")
    max_hold_attempts = traits.Int(5, desc='Number of attempts to initiate a trial before skipping to the next one')
    tracking_reward_time = traits.Float(.15, desc="Length of reward for tracking in")
    tracking_reward_interval = traits.Float(.5, desc="Length of time between reward for tracking in (must be greater than reward time)")
    
    def init(self):
        self.trial_dtype = np.dtype([('trial', 'u4'), ('index', 'u4'), ('target', 'f8',(3,)), ('disturbance', 'f8',(3,)), ('is_disturbance', '?')])
        super().init()

        self.frame_index = 0 # index in the frame in a trial
        self.total_distance_error = 0 # Euclidian distance between cursor and target during each trial
        self.trial_timed_out = True # check if the trial is finished
        self.plant_position = []
        self.disturbance_trial = False
        self.repeat_freq_set = False
        self.gen_index = -1

        if self.velocity_control:
            print('VELOCITY CONTROL')
        else:
            print('POSITION CONTROL') # default is position control - see manualcontrolmixin
        self.pos_offset = [0,0,0]
        self.vel_offset = [0,0,0]

    def _parse_next_trial(self):
        '''Get the required data from the generator'''
        # yield idx, pts, disturbance, dis_trajectory, sample_rate :
        self.gen_index, self.targs, self.disturbance_trial, self.disturbance_path, self.sample_rate = self.next_trial # targs and disturbance are same length

        self.targs = np.squeeze(self.targs,axis=0)
        self.disturbance_path = np.squeeze(self.disturbance_path)

        WIDTH, HEIGHT = self.window_size[0], self.window_size[1]
        lookahead = np.zeros((self.lookahead,np.shape(self.targs)[1])) # (30,3)

        self.targs = self.trajectory_amplitude*self.targs
        self.disturbance_path = self.disturbance_amplitude*self.disturbance_path
        # print(np.amax(self.targs), np.amax(self.disturbance_path))

        self.targs = np.concatenate((lookahead, self.targs),axis=0) # (time_length*sample_rate+30,3) # targs and disturbance are no longer same length
    
    def tracking_task_start_wait(self):
        print(self.gen_index)

        self.trial_record['trial'] = self.calc_trial_num()
        self.trial_record['index'] = self.gen_index
        self.trial_record['is_disturbance'] = self.disturbance_trial
        for i in range(len(self.disturbance_path)):
            # Update the data sinks with trial information --> bmi3d_trials
            self.trial_record['target'] = self.targs[i+self.lookahead]
            self.trial_record['disturbance'] = self.disturbance_path[i]
            self.sinks.send("trials", self.trial_record)

        # trial is not finished
        self.trial_timed_out = False

        # number of times this trajectory has been attempted
        self.tries = 0

        # index into trajectory
        self.frame_index = -1

        # number of frames in trajectory
        '''Nothing generic to do.'''
        self.trajectory_length = len(self.targs)

        # saved plant poitions
        self.plant_position = []

    def _start_wait(self):
        # Call parent method to draw the next target capture sequence from the generator
        super()._start_wait()
        if self.sample_rate != self.fps:
            print('WARNING: generator sample rate should equal FSM fps!')

        if self.repeat_freq_set:
            try:
                self.next_trial = next(self.gen)
                self._parse_next_trial()
            except StopIteration:
                self.end_task()

        self.tracking_task_start_wait()

    def _while_wait(self):
        '''Nothing generic to do.'''
        pass

    def _start_wait_retry(self):
        self.tracking_task_start_wait()

    def _while_wait_retry(self):
        '''Nothing generic to do.'''
        pass

    def _start_trajectory(self):
        self.tries += 1
        self.frame_index += 1

    def _while_trajectory(self):
        '''Nothing generic to do.'''
        pass        

    def _end_trajectory(self):
        '''Nothing generic to do.'''
        pass

    def _start_hold(self):
        '''Nothing generic to do.'''
        pass

    def _while_hold(self):
        self.pos_offset = [0,0,0]
        self.vel_offset = [0,0,0]

    def _end_hold(self):
        '''Nothing generic to do.'''
        pass

    def _start_tracking_in(self):
        '''Nothing generic to do.'''
        pass

    def _while_tracking_in(self):
        '''Nothing generic to do.'''
        pass

    def _end_tracking_in(self):
        '''Nothing generic to do.'''
        pass

    def _start_tracking_out(self):
        '''Nothing generic to do.'''
        pass

    def _while_tracking_out(self):
        '''Nothing generic to do.'''
        pass

    def _end_tracking_out(self):
        '''Nothing generic to do.'''
        pass

    def _start_timeout_penalty(self):
        '''Nothing generic to do.'''
        # self._increment_tries()
        pass

    def _while_timeout_penalty(self):
        self.pos_offset = [0,0,0]
        self.vel_offset = [0,0,0]

    def _end_timeout_penalty(self):
        '''Nothing generic to do.'''
        pass

    def _start_hold_penalty(self):
        '''Nothing generic to do.'''
        # self._increment_tries()
        pass

    def _while_hold_penalty(self):
        self.pos_offset = [0,0,0]
        self.vel_offset = [0,0,0]

    def _end_hold_penalty(self):
        '''Nothing generic to do.'''
        pass

    def _start_tracking_out_penalty(self):
        '''Nothing generic to do.'''
        pass

    def _while_tracking_out_penalty(self):
        self.pos_offset = [0,0,0]
        self.vel_offset = [0,0,0]

    def _end_tracking_out_penalty(self):
        '''Nothing generic to do.'''
        pass

    def _start_reward(self):
        '''Nothing generic to do.'''
        pass

    def _while_reward(self):
        self.pos_offset = [0,0,0]
        self.vel_offset = [0,0,0]

    def _end_reward(self):
        '''Nothing generic to do.'''
        pass

    def _start_pause(self):
        '''Nothing generic to do.'''
        pass

    def _while_pause(self):
        '''Nothing generic to do.'''
        self.pos_offset = [0,0,0]
        self.vel_offset = [0,0,0]

    def _end_pause(self):
        '''Nothing generic to do.'''
        pass

    ################## State transition test functions ##################
    def _test_start_trial(self, time_in_state):
        '''Start next trial automatically. You may want this to instead be
            - a random delay
            - require some initiation action
        '''
        return time_in_state > self.wait_time

    def _test_timeout(self, time_in_state):
        return time_in_state > self.timeout_time

    def _test_hold_complete(self, time_in_state):
        '''
        Test whether the target is held long enough to declare the
        trial a success

        Possible options
            - Target held for the minimum requred time (implemented here)
            - Sensorized object moved by a certain amount
            - Sensorized object moved to the required location
            - Manually triggered by experimenter
        '''
        return time_in_state > self.hold_time

    def _test_trial_complete(self, time_in_state):
        '''Test whether the trajectory is finished'''
        return self.frame_index + self.lookahead == self.trajectory_length

    def _test_tracking_out_timeout(self, time_in_state):
        return time_in_state > self.tracking_out_time

    def _test_timeout_penalty_end(self, time_in_state):
        return time_in_state > self.timeout_penalty_time #or self.pause

    def _test_hold_penalty_end(self, time_in_state):
        return (time_in_state > self.hold_penalty_time) and (self.tries==self.max_hold_attempts) #or self.pause

    def _test_hold_penalty_end_retry(self, time_in_state):
        return (time_in_state > self.hold_penalty_time) and (self.tries<self.max_hold_attempts) #or self.pause

    def _test_tracking_out_penalty_end(self, time_in_state):
        return time_in_state > self.tracking_out_penalty_time #or self.pause

    def _test_reward_end(self, time_in_state):
        return time_in_state > self.reward_time

    def _test_enter_target(self, time_in_state):
        '''This function is task-specific and not much can be done generically'''
        return False

    def _test_leave_target(self, time_in_state):
        '''This function is task-specific and not much can be done generically'''
        return False

    def _test_start_pause(self, time_in_state):
        return self.pause

    def _test_end_pause(self, time_in_state):
        return not self.pause
    
    def update_report_stats(self):
        '''
        see experiment.Experiment.update_report_stats for docs
        '''
        super().update_report_stats()
        self.reportstats['Trial #'] = self.calc_trial_num()
        self.reportstats['Reward/min'] = np.round(self.calc_events_per_min('reward', 120.), decimals=2)
        

class ScreenTargetTracking(TargetTracking, Window):
    """Concrete implementation of Target Tracking task where the target is moving and
    are tracked by holding the cursor within the moving target"""

    limit2d = traits.Bool(True, desc="Limit cursor movement to 2D")
    limit1d = traits.Bool(True, desc="Limit cursor movement to 1D")

    sequence_generators = [
        'tracking_target_chain', 'tracking_target_debug', 'tracking_target_training'
    ]

    hidden_traits = ['cursor_color', 'trajectory_color', 'cursor_bounds', 'cursor_radius', 'plant_hide_rate', 'starting_pos']
    targets = []

    is_bmi_seed = True

    # Runtime settable traits
    target_radius = traits.Float(2, desc="Radius of targets in cm") #2,0.75
    trajectory_radius = traits.Float(.5, desc="Radius of targets in cm")
    trajectory_color = traits.OptionsList("gold", *target_colors, desc="Color of the trajectory", bmi3d_input_options=list(target_colors.keys()))
    target_color = traits.OptionsList("yellow", *target_colors, desc="Color of the target", bmi3d_input_options=list(target_colors.keys()))
    plant_hide_rate = traits.Float(0.0, desc='If the plant is visible, specifies a percentage of trials where it will be hidden')
    plant_type = traits.OptionsList(*plantlist, bmi3d_input_options=list(plantlist.keys()))
    plant_visible = traits.Bool(True, desc='Specifies whether entire plant is displayed or just endpoint')
    cursor_radius = traits.Float(.5, desc='Radius of cursor in cm')
    cursor_color = traits.OptionsList("pink", *target_colors, desc='Color of cursor endpoint', bmi3d_input_options=list(target_colors.keys()))
    cursor_bounds = traits.Tuple((-10., 10., 0., 0., -10., 10.), desc='(x min, x max, z min, z max, y min, y max)')
    starting_pos = traits.Tuple((5., 0., 5.), desc='Where to initialize the cursor')
    fps = traits.Float(60, desc="Rate at which the FSM is called in Hz") # originally set by class Experiment
    trajectory_amplitude = traits.Float(1, desc='Scale factor applied to the trajectory')
    disturbance_amplitude = traits.Float(1, desc='Scale factors applied to the disturbance')
    always_1d = traits.Bool(False, desc='always constrain movement to 1d')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Initialize the plant
        if not hasattr(self, 'plant'):
            self.plant = plantlist[self.plant_type]
        self.plant.set_bounds(np.array(self.cursor_bounds))
        self.plant.set_color(target_colors[self.cursor_color])
        self.plant.set_cursor_radius(self.cursor_radius)
        self.plant_vis_prev = True
        self.cursor_vis_prev = True
        self.lookahead = 30 # number of frames to create a "lookahead" window of 0.5 seconds (half the screen)
        self.original_limit1d = self.limit1d # keep track of original settable trait
        
        if not self.always_1d:
            self.limit1d = False # allow 2d movement before center-hold initiation
        
        # Add graphics models for the plant and targets to the window
        if hasattr(self.plant, 'graphics_models'):
            for model in self.plant.graphics_models:
                self.add_model(model)

        # Instantiate the targets
        instantiate_targets = kwargs.pop('instantiate_targets', True)
        if instantiate_targets:
            # This is the center target being followed by the user
            self.target = VirtualCircularTarget(target_radius=self.target_radius, target_color=target_colors[self.target_color])

            # This is the trajectory that spans the screen
            self.trajectory = VirtualCableTarget(target_radius=self.trajectory_radius, target_color=target_colors[self.trajectory_color])

            # This is the progress bar
            self.bar = VirtualRectangularTarget(target_width=1, target_height=0, target_color=(0., 1., 0., 0.75), starting_pos=[0,0,9])
            # print('INIT TRAJ')

        # Declare any plant attributes which must be saved to the HDF file at the _cycle rate
        for attr in self.plant.hdf_attrs:
            self.add_dtype(*attr)

    def init(self):
        self.add_dtype('trial', 'u4', (1,))
        self.add_dtype('gen_idx', 'int', (1,)) # dtype needs to be able to represent -1
        self.add_dtype('plant_visible', '?', (1,))
        self.add_dtype('current_target', 'f8', (3,))
        self.add_dtype('current_disturbance', 'f8', (3,)) # see task_data['manual_input'] for cursor position without added disturbance
        self.add_dtype('current_target_validate', 'f8', (3,))
        super().init()
        self.plant.set_endpoint_pos(np.array(self.starting_pos))

    def _cycle(self):
        '''
        Calls any update functions necessary and redraws screen
        '''
        self.move_effector(pos_offset=np.asarray(self.pos_offset), vel_offset=np.asarray(self.vel_offset))

        # Run graphics commands to show/hide the plant if the visibility has changed
        self.update_plant_visibility()
        self.task_data['plant_visible'] = self.plant_visible

        # Save plant status to HDF file
        plant_data = self.plant.get_data_to_save() # for cursor plant, save cursor position
        for key in plant_data:
            self.task_data[key] = plant_data[key] # task_data['cursor']

        # Update the trial index
        self.task_data['trial'] = self.calc_trial_num()
        self.task_data['gen_idx'] = self.gen_index
        # print(self.task_data['gen_idx'])
        
        # Save the target position at each cycle. 
        if self.trial_timed_out:
             self.task_data['current_target'] = [np.nan,np.nan,np.nan]
             self.task_data['current_disturbance'] = [np.nan,np.nan,np.nan]
             self.task_data['current_target_validate'] = self.target.get_position() # default VirtualCircularTarget position is [0,0,0]
        else:
            self.task_data['current_target'] = self.targs[self.frame_index+self.lookahead]
            self.task_data['current_disturbance'] = self.disturbance_path[self.frame_index]
            self.task_data['current_target_validate'] = self.target.get_position()

        super()._cycle()

    def move_effector(self):
        '''Move the end effector, if a robot or similar is being controlled'''
        pass

    def run(self):
        '''
        See experiment.Experiment.run for trajectorydocumentation.
        '''
        # Fire up the plant. For virtual/simulation plants, this does little/nothing.
        self.plant.start()
        
        # Include some cleanup in case the parent class has errors
        try:
            super().run()
        finally:
            self.plant.stop()

    ##### HELPER AND UPDATE FUNCTIONS ####
    def update_plant_visibility(self):
        ''' Update plant visibility'''
        if self.plant_visible != self.plant_vis_prev:
            self.plant_vis_prev = self.plant_visible
            self.plant.set_visibility(self.plant_visible)

    def update_frame(self):
        self.target.move_to_position(np.array([0,0,self.targs[self.frame_index+self.lookahead][2]])) # xzy
        self.trajectory.move_to_position(np.array([-self.frame_index/3,0,0])) # same update constant works for 60 and 120 hz
        self.target.show()
        self.trajectory.show()
        self.frame_index +=1

    #### TEST FUNCTIONS ####
    def _test_enter_target(self, time_in_state):
        '''
        Test if the cursor is inside the center target 
        '''
        cursor_pos = self.plant.get_endpoint_pos()
        d = np.linalg.norm(cursor_pos - self.target.get_position())
        return d <= (self.target_radius - self.cursor_radius) or super()._test_enter_target(time_in_state) or (self.hold_time==0 and self.state=='trajectory')

    def _test_leave_target(self, time_in_state):
        '''
        Test if the cursor is outside the center target
        '''
        cursor_pos = self.plant.get_endpoint_pos()
        d = np.linalg.norm(cursor_pos - self.target.get_position())
        return d > (self.target_radius - self.cursor_radius) or super()._test_leave_target(time_in_state)

    #### STATE FUNCTIONS ####
    def setup_start_wait(self):
        if self.calc_trial_num() == 0:
            # Instantiate the targets here so they don't show up in any states that might come before "wait" 
            for model in self.target.graphics_models:
                self.add_model(model)
                self.target.hide()
                
            for model in self.trajectory.graphics_models:
                self.add_model(model)
                self.trajectory.hide()

            for model in self.bar.graphics_models:
                self.add_model(model)
                self.bar.hide()

        # Allow 2d movement
        if not self.always_1d:
            self.limit1d = False

        # Set up for progress bar
        self.bar_width = 12        
        self.tracking_frame_index = 0
        
        # Set up the next trajectory
        next_trajectory = np.array(np.squeeze(self.targs)[:,2])
        next_trajectory[:self.lookahead] = next_trajectory[self.lookahead]

        if hasattr(self, 'trajectory'):
            for model in self.trajectory.graphics_models:
                self.remove_model(model)
            del self.trajectory

        self.trajectory = VirtualCableTarget(target_radius=self.trajectory_radius, target_color=target_colors[self.trajectory_color], trajectory=next_trajectory)

        for model in self.trajectory.graphics_models:
                self.add_model(model)

        self.target.hide()
        self.trajectory.hide()
        # self.in_end_state = False      

    def setup_screen_reset(self):
        # Hide target and trajectory
        self.target.hide()
        self.target.reset()
        self.trajectory.hide()
        self.trajectory.reset()
        self.bar.hide()
        self.bar.reset()

    def _start_wait(self):
        super()._start_wait()
        self.setup_start_wait()
        print('WAIT')

    def _while_wait(self):
        super()._while_wait()

    def _start_wait_retry(self):
        super()._start_wait_retry()
        self.setup_start_wait()
        # print('WAIT RETRY')

    def _while_wait_retry(self):
        super()._while_wait_retry()
        # # Add disturbance

    def _start_trajectory(self):
        super()._start_trajectory()
        if self.frame_index == 0:
            self.target.move_to_position(np.array([0,0,self.targs[self.frame_index+self.lookahead][2]])) # tablet screen x-axis ranges -19,19, center 0
            self.trajectory.move_to_position(np.array([0,0,0])) # tablet screen x-axis ranges 0,41.33333, center 22ish
            # print(self.target.get_position())
            # print(self.trajectory.get_position())

            self.target.show()
            self.trajectory.show()
            # print('SHOW TRAJ')
            self.sync_event('TARGET_ON')

    def _while_trajectory(self):
        super()._while_trajectory()
        # # Add disturbance

    def _start_hold(self):
        super()._start_hold()
        # print('START HOLD')
        self.sync_event('TRIAL_START')
        # Cue successful tracking
        self.target.cue_trial_end_success()

    def _while_hold(self):
        super()._while_hold()
        # # Add disturbance

    def _start_tracking_in(self):
        super()._start_tracking_in()
        # print('START TRACKING')
        self.sync_event('CURSOR_ENTER_TARGET')
        # Revert to settable trait
        self.limit1d = self.original_limit1d
        # Cue successful tracking
        self.target.cue_trial_end_success()

    def _while_tracking_in(self):
        super()._while_tracking_in()
    
        # Add disturbance
        cursor_pos = self.plant.get_endpoint_pos()
        if self.disturbance_trial == True:
            if self.velocity_control:
                # TODO check manualcontrolmixin for how to implement velocity control
                self.vel_offset = (cursor_pos + self.disturbance_path[self.frame_index])*1/self.fps
            else: 
                # position control
                self.pos_offset = self.disturbance_path[self.frame_index]

        # Move target and trajectory to next frame so it appears to be moving
        self.update_frame()

        # Check if the trial is over and there are no more target frames to display
        if self.frame_index+self.lookahead >= np.shape(self.targs)[0]:
            self.trial_timed_out = True
            

    def _start_tracking_out(self):
        super()._start_tracking_out()
        # print('STOP TRACKING')
        self.sync_event('CURSOR_LEAVE_TARGET')
        # Reset target color
        self.target.reset()

    def _while_tracking_out(self):
        super()._while_tracking_out()

        # Add disturbance
        cursor_pos = self.plant.get_endpoint_pos()
        if self.disturbance_trial == True:
            if self.velocity_control:
                # TODO check manualcontrolmixin for how to implement velocity control
                self.vel_offset = (cursor_pos + self.disturbance_path[self.frame_index])*1/self.fps
            else: 
                # position control
                self.pos_offset = self.disturbance_path[self.frame_index]

        # Move target and trajectory to next frame so it appears to be moving
        self.update_frame()

        # Check if the trial is over and there are no more target frames to display
        if self.frame_index+self.lookahead >= np.shape(self.targs)[0]:
            self.trial_timed_out = True

    def _start_timeout_penalty(self):
        super()._start_timeout_penalty()
        # print('START TIMEOUT')
        self.sync_event('TIMEOUT_PENALTY')

        # self.in_end_state = True
        self.setup_screen_reset()

        # skip to next generated trial using same freq set
        self.repeat_freq_set = True

    def _while_timeout_penalty(self):
        super()._while_timeout_penalty()
        # # Add disturbance

    def _end_timeout_penalty(self):
        super()._end_timeout_penalty()
        self.sync_event('TRIAL_END')
            
    def _start_hold_penalty(self):
        super()._start_hold_penalty()
        # print('START HOLD TIMEOUT')
        self.sync_event('HOLD_PENALTY')

        # self.in_end_state = True
        self.setup_screen_reset()

        # skip to next generated trial using same freq set
        self.repeat_freq_set = True

    def _while_hold_penalty(self):
        super()._while_hold_penalty()
        # # Add disturbance

    def _end_hold_penalty(self):
        super()._end_hold_penalty()
        # print('TRIAL END')
        self.sync_event('TRIAL_END')

    def _start_tracking_out_penalty(self):
        super()._start_tracking_out_penalty()
        # print('START TRACKING TIMEOUT')
        self.sync_event('OTHER_PENALTY')
        # self.in_end_state = True
        # Cue failed trial
        self.target.cue_trial_end_failure()

        # skip to next generated trial using same freq set
        self.repeat_freq_set = True

    def _while_tracking_out_penalty(self):
        super()._while_tracking_out_penalty()
        # # Add disturbance

    def _end_tracking_out_penalty(self):
        super()._end_tracking_out_penalty()
        # print('TRIAL END')
        self.sync_event('TRIAL_END')
        self.setup_screen_reset()

    def _start_reward(self):
        super()._start_reward()
        # print('REWARD')
        self.sync_event('REWARD')
        # self.in_end_state = True
        # Cue successful trial
        self.target.cue_trial_end_success()
        self.reward_frame_index = 0

        # use next generated trial using other freq set
        self.repeat_freq_set = False

    def _while_reward(self):
        super()._while_reward()
        # # Add disturbance

    def _end_reward(self):
        super()._end_reward()
        # print('TRIAL END')
        self.sync_event('TRIAL_END')
        self.setup_screen_reset()

    def _start_pause(self):
        super()._start_pause()
        # print('START PAUSE')
        self.sync_event('PAUSE_START') # self.sync_event('PAUSE_START') - this will overwrite TRIAL_END whenever you pause during an end state
        self.setup_screen_reset()

    def _while_pause(self):
        super()._while_pause()

    def _end_pause(self):
        super()._end_pause()
        # print('END PAUSE')
        self.sync_event('PAUSE_END')

    @staticmethod
    def calc_sum_of_sines(times, frequencies, amplitudes, phase_shifts):
        '''
        Generates the trajectories for the experiment
        '''
        t = times
        t = np.asarray(t).copy(); t.shape = (t.size,1)

        f = frequencies
        f = f.copy()
        f = np.reshape(f, (1,f.size))

        a = amplitudes
        a = a.copy()
        a = np.reshape(a, (1,a.size))

        p = phase_shifts        
        p = p.copy()
        p = np.reshape(p, (1,p.size))

        assert f.shape == a.shape == p.shape,"Shape of frequencies, amplitudes, and phase shifts must match"

        o = np.ones(t.shape)
        trajectory = np.sum(np.dot(o,a) * np.sin(2*np.pi*(np.dot(t,f) + np.dot(o,p))),axis=1)
        A = np.sum(np.dot(o,a)[0])
        
        return trajectory, A

    @staticmethod
    def calc_sum_of_sines_ramp(times, ramp, ramp_down, frequencies, amplitudes, phase_shifts):
            '''
            Adds a 1/t ramp up and ramp down at the start and end so the trajectories start and end at zero.
            '''
            t = times
            t = np.asarray(t).copy(); t.shape = (t.size,1)

            r = ramp
            rd = ramp_down

            trajectory, A = ScreenTargetTracking.calc_sum_of_sines(t, frequencies, amplitudes, phase_shifts)

            if r > 0:
                trajectory *= (( t*(t <= r)/r + (t > r) ).flatten())**2
                # trajectory *= (((t*(t <= r)/r) + ((t > r) & (t < (t[-1]-r))) + ((t[-1]-t)*(t >= (t[-1]-r))/r)).flatten())**2

            if rd > 0:
                trajectory *= (( (t < (t[-1]-rd)) + ((t[-1]-t)*(t >= (t[-1]-rd))/rd) ).flatten())**2

            return trajectory, A

    @staticmethod
    def generate_trajectories(num_trials=2, time_length=20, seed=40, sample_rate=120, base_period=20, ramp=0, ramp_down=0, num_primes=8):
        '''
        Sets up variables and uses prime numbers to call the above functions and generate then trajectories
        ramp is time length for preparatory lines
        '''
        np.random.seed(seed)
        hz = sample_rate # Hz -- sampling rate
        dt = 1/hz # sec -- sampling period

        T0 = base_period # sec -- base period
        w0 = 1./T0 # Hz -- base frequency

        r = ramp # "ramp up" duration (see sum_of_sines_ramp)
        rd = ramp_down # "ramp down" duration (see sum_of_sines_ramp)
        P = time_length/T0 # number of periods in signal
        T = P*T0+r+rd # sec -- signal duration
        dw = 1./T # Hz -- frequency resolution
        W = 1./dt/2 # Hz -- signal bandwidth

        full_primes = np.asarray([2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41, 43, 47, 53, 59, 61, 67, 71, 73, 79, 83, 89, 97, 
            101, 103, 107, 109, 113, 127, 131, 137, 139, 149, 151, 157, 163, 167, 173, 179, 181, 191, 193, 197, 199])
        primes = full_primes[:num_primes]
        # primes_ind = np.where(full_primes <= T0)
        # primes = full_primes[primes_ind]

        f = primes*w0 # stimulated frequencies
        f_ref = f.copy()
        f_dis = f.copy()

        a = 1/(1+np.arange(f.size)) # amplitude
        a_ref = a.copy()
        a_dis = a.copy()

        o = np.random.rand(num_trials,primes.size) # phase offset
        o_ref = o.copy()
        o_dis = o.copy()*0.8

        t = np.arange(0,T,dt) # time samplesseed
        w = np.arange(0,W,dw) # frequency samples

        N = t.size # = T/dt -- number of samples
        
        # create trials dictionary
        trials = dict(
            id=np.arange(num_trials), times=np.tile(t,(num_trials,1)), ref=np.zeros((num_trials,N)), dis=np.zeros((num_trials,N))
            )

        # randomize order of first two trials to provide random starting point
        order = np.random.choice([0,1])
        if order == 0:
            trial_order = [(1,'E','O'),(1,'O','E')]
        elif order == 1:
            trial_order = [(1,'O','E'),(1,'E','O')]

        # generate reference and disturbance trajectories for all trials
        for trial_id, (num_reps,ref_ind,dis_ind) in enumerate(trial_order*int(num_trials/2)):   
            if ref_ind == 'E': 
                sines_r = np.arange(len(primes))[0::2] # use even indices
            elif ref_ind == 'O': 
                sines_r = np.arange(len(primes))[1::2] # use odd indices
            else:
                sines_r = np.arange(len(primes))
            if dis_ind == 'E':
                sines_d = np.arange(len(primes))[0::2]
            elif dis_ind == 'O':
                sines_d = np.arange(len(primes))[1::2]
            else:
                sines_d = np.arange(len(primes))
            
            ref_trajectory, ref_A = ScreenTargetTracking.calc_sum_of_sines_ramp(t, r, rd, f_ref[sines_r], a_ref[sines_r], o_ref[trial_id][sines_r])
            dis_trajectory, dis_A = ScreenTargetTracking.calc_sum_of_sines_ramp(t, r, rd, f_dis[sines_d], a_dis[sines_d], o_dis[trial_id][sines_d])
            
            # normalized trajectories
            trials['ref'][trial_id] = ref_trajectory/ref_A   # previously, denominator was np.sum(a_ref)
            trials['dis'][trial_id] = dis_trajectory/dis_A   # previously, denominator was np.sum(a_dis)
            # print(trial_order, ref_A, dis_A)
        
        return trials, trial_order

    @staticmethod
    def generate_trajectory(primes, base_period, ramp = .0):
        '''
        Sets up variables and uses prime numbers to call the above functions and generate then trajectories
        ramp is time length for preparatory lines
        '''
        hz = 60 # Hz -- sampling rate
        dt = 1/hz # sec -- sampling period

        T0 = base_period # sec -- base period
        w0 = 1/T0 # Hz -- base frequency

        P = 1 # number of periods in signal
        T = P*T0 # sec -- signal duration
        r = ramp # "ramp" duration (see sum_of_sines_ramp)
        dw = 1/T # Hz -- frequency resolution
        W = dw*T/dt/2 # Hz -- signal bandwidth

        f = primes # stimulated frequencies
        a = 1/(1+np.arange(f.size)) # amplitude
        o = 2*np.pi*np.random.rand(primes.size) # phase offset

        t = np.arange(0,T,dt) # time samples
        w = np.arange(0,W,dw) # frequency samples

        N = t.size # = T/dt -- number of samples

        #trajectory = ScreenTargetTracking.calc_sum_of_sines_ramp(t, r, f, a, o/(2*np.pi))
        trajectory = ScreenTargetTracking.calc_sum_of_sines_ramp(t, r, f, a, o)
        normalized_trajectory = trajectory/np.sum(a)
        return normalized_trajectory
    
    ### Generator functions ####
    @staticmethod
    def tracking_target_chain(nblocks=1, ntrials=2, time_length=20, ramp=0, ramp_down=0, num_primes=8, seed=40, sample_rate=120, disturbance=True, boundaries=(-10,10,-10,10)):
        '''
        Generates a sequence of 1D (z axis) target trajectories

        Parameters
        ----------
        nblocks : int
            The number of blocks in the session
        ntrials : int
            The number of trials in a block
        time_length : int
            The length of one trial in seconds
        seed : int
            The seed for the random generator
        sample_rate : int
            The sample rate of the generated trajectories
        ramp : float
            The length of ramp up into a trial in seconds
        disturbance : boolean
            Whether to add disturbance to the cursor (disturbance is generated regardless)
        boundaries: 4 element tuple
            The limits of the allowed target locations (-x, x, -z, z)

        Returns
        -------
        idx : [nblocks*ntrials x 1] array of trial indices
        pts : [nblocks*ntrials x 1] array of 3D target coordinates
        disturbance : boolean
        dis_trajectory : [nblocks*ntrials x 1] array of 3D disturbance coordinates
        '''
        idx = 0
        base_period = 20
        for block_id in range(nblocks):                
            trials, trial_order = ScreenTargetTracking.generate_trajectories(
                num_trials=ntrials, time_length=time_length, seed=seed, sample_rate=sample_rate, base_period=base_period, ramp=ramp, ramp_down=ramp_down, num_primes=num_primes
                )
            for trial_id in range(ntrials):
                pts = []
                ref_trajectory = np.zeros((int((time_length+ramp+ramp_down)*sample_rate),3))
                dis_trajectory = ref_trajectory.copy()
                ref_trajectory[:,2] = trials['ref'][trial_id]
                dis_trajectory[:,2] = trials['dis'][trial_id] # scale will determine lower limit of target size for perfect tracking
                pts.append(ref_trajectory)
                yield idx, pts, disturbance, dis_trajectory, sample_rate
                idx += 1

    @staticmethod
    def tracking_target_debug(nblocks=1, ntrials=2, time_length=20, seed=40, sample_rate=60, ramp=0, disturbance=True, boundaries=(-10,10,-10,10)):
        '''
        Generates a sequence of 1D (z axis) target trajectories for debugging
        '''
        idx = 0
        base_period = 20
        for block_id in range(nblocks):                
            trials, trial_order = ScreenTargetTracking.generate_trajectories(
                num_trials=ntrials, time_length=time_length, seed=seed, sample_rate=sample_rate, base_period=base_period, ramp=ramp
                )
            for trial_id in range(ntrials):
                # if trial_id==0:
                #     plt.figure(); plt.plot(trials['times'][trial_id],trials['ref'][trial_id])
                #     plt.plot(trials['times'][trial_id],trials['dis'][trial_id])
                #     plt.show()
                pts = []
                ref_trajectory = np.zeros((int((time_length+ramp)*sample_rate),3))
                dis_trajectory = ref_trajectory.copy()
                ref_trajectory[:,2] = trials['ref'][trial_id]
                dis_trajectory[:,2] = trials['dis'][trial_id] # scale will determine lower limit of target size for perfect tracking
                pts.append(ref_trajectory)
                yield idx, pts, disturbance, dis_trajectory, sample_rate
                idx += 1
    
    @staticmethod
    def tracking_target_training(nblocks=1, ntrials=2, time_length=5, frequencies = [1,.75], boundaries=(-10,10,-10,10)):
        '''
        Generates a sequence of 1D (z axis) target trajectories for training

        Parameters
        ----------
        nblocks : int
            The number of tracking trials in the sequence.
        ntrials : int
            The number trials in a block
        time_length : int
        frequencies: numpy.ndarray
            A list of frequencies used to generate the trajectories
        boundaries: 4 element Tuple
            The limits of the allowed target locations (-x, x, -z, z)

        Returns
        -------
        [nblocks*ntrials x 1] array of tuples containing trial indices and [time_length*60 x 3] target coordinates
        '''
        idx = 0
        buffer_space_bef = int(60*1.3) # 1.3 seconds of straight line before trial #78 frames
        buffer_space_aft = int(60*1.5) # 1.5 seconds of straight line before trial #90 frames
        frames = int(np.round(time_length*60)) #300 frames when time_length=5
        y_primes_freq = np.array(frequencies)

        for i in range(nblocks):
            for j in range(ntrials):
                
                disturbance = False
                disturbance_path = np.zeros((frames+buffer_space_bef+buffer_space_aft,1))
                trajectory = np.zeros((frames+buffer_space_bef+buffer_space_aft,3))
                sum_of_sins_path = ScreenTargetTracking.generate_trajectory(y_primes_freq,time_length)
                pts = []
                trajectory[:,2] = 5*np.concatenate((sum_of_sins_path[0]*np.ones(buffer_space_bef),sum_of_sins_path,sum_of_sins_path[-1]*np.ones(buffer_space_aft)))
                pts.append(trajectory)
                yield idx, pts, disturbance, disturbance_path, None
                idx += 1


import time
from riglib import experiment
from analysis import online_analysis
from features.debug_features import OnlineAnalysis, ReplayCursor, ReplayEye
from features.peripheral_device_features import MouseControl
from built_in_tasks.manualcontrolmultitasks import ManualControl
from riglib.stereo_opengl.window import Window2D
import unittest
import numpy as np
import os
import threading

class TestOnlineAnalysis(unittest.TestCase):

    @unittest.skip("")
    def test_manual_run(self):

        analysis = online_analysis.OnlineDataServer('localhost', 5000)

        Exp = experiment.make(experiment.Experiment, feats=[OnlineAnalysis])
        exp = Exp(fps=1, session_length=5, online_analysis_ip='localhost', online_analysis_port=5000)
        print(exp.dtype)
        exp.init()

        t0 = time.time()
        while time.time() - t0 < 2:
            analysis.update()
        print('done updating')
        self.assertEqual(analysis.task_params['experiment_name'], 'Experiment')

        threading.Thread(target=exp.run).start()
        time.sleep(1)

        while True:
            if not analysis.update():
                break
        self.assertTrue(analysis.is_running)
        self.assertFalse(analysis.is_completed)
        self.assertEqual(analysis.state, 'wait')

        time.sleep(6)
        while True:
            if not analysis.update():
                break
        self.assertTrue(analysis.is_completed)
        self.assertFalse(analysis.is_running)
        self.assertEqual(analysis.state, None)

        analysis._stop()

    @unittest.skip("")
    def test_threaded(self):

        analysis = online_analysis.OnlineDataServer('localhost', 5000)

        # Start exp 1
        Exp = experiment.make(experiment.Experiment, feats=[OnlineAnalysis])
        exp = Exp(fps=10, session_length=18, online_analysis_ip='localhost', online_analysis_port=5000)
        print(exp.dtype)
        exp.init()

        # Start analysis
        analysis.start()
        time.sleep(1)
        exp.run()

        # Wrap up
        analysis.stop()
        analysis.join()

    def test_cursor(self):

        analysis = online_analysis.OnlineDataServer('localhost', 5000)

        # Load replay data
        import tables
        test_dir = os.path.dirname(os.path.abspath(__file__))
        hdf_filepath = os.path.join(test_dir, 'test_data/beig20240419_21_te16793.hdf')
        with tables.open_file(hdf_filepath, 'r') as f:
            task = f.root.task.read()
            trial = f.root.trials.read()
        targets = []
        for trial in trial:
            trial_num = trial['trial']
            if len(targets) <= trial_num:
                targets.append(([],[]))
            targets[trial_num][0].append(trial['index'])
            targets[trial_num][1].append(trial['target'])
        cursor = task['cursor']
        eye = task['eye']

        # Start exp 1
        os.environ['DISPLAY'] = ':0.0'
        seq = targets
        Exp = experiment.make(ManualControl, feats=[Window2D, ReplayCursor, ReplayEye, OnlineAnalysis])
        exp = Exp(seq, fps=120, session_length=28, online_analysis_ip='localhost', online_analysis_port=5000,
                  fullscreen=False, window_size=(800,600), replay_cursor_data=cursor,
                  replay_eye_data=eye, wait_time=0)

        print(exp.dtype)
        exp.init()

        # Start analysis
        analysis.start()
        exp.run()

        # Wrap up
        analysis.stop()
        analysis.join()



if __name__ == '__main__':
    unittest.main()


'''
Force-based size task
'''
import numpy as np

from riglib.experiment import traits, LogExperiment
from riglib.stereo_opengl.window import Window
from .target_graphics import *

disk_colors = {
    'target': (0.4, 0.4, 0.4, 0.75),
    'target_bright': (0.5, 0.5, 0.5, 0.75),
    'cursor': (1, 1, 1, 0.5),
}

class DiskMatchingMixin(traits.HasTraits):
    """
    Adds an annular target and disk cursor. To be combined with manual control experiments.
    """
 
    # Runtime settable traits
    disk_cursor_offset = traits.Float(0., desc="offset from raw joystick value to disk cursor radius")
    disk_cursor_gain = traits.Float(50., desc="conversion from raw joystick value to disk cursor radius")
    disk_cursor_bounds = traits.Tuple((0.5, 8.), desc="(min, max) radius (in cm) of the disk cursor")
    disk_target_radius = traits.Tuple((2., 5.), desc="(min, max) inner radius (in cm) of the disk target")
    disk_target_tolerance = traits.Float(0.5, desc="distance from inner to outer radius (in cm) of disk target")
    disk_target_color = traits.OptionsList("target", *disk_colors, desc="Color of the disk target", bmi3d_input_options=list(disk_colors.keys()))
    disk_cursor_color = traits.OptionsList("cursor", *disk_colors, desc='Color of disk cursor ', bmi3d_input_options=list(disk_colors.keys()))
    disk_pos = traits.Tuple((0., 0., 0.), desc='Where to locate the disks') 

    def __init__(self, *args, **kwargs):
        super().__init__(self, *args, **kwargs)
        self.disk_cursor = VirtualCircularTarget(target_radius=0, target_color=disk_colors[self.disk_cursor_color], starting_pos=self.disk_pos)
        self.disk_target = VirtualTorusTarget() # just a placeholder
        for model in self.disk_cursor.graphics_models:
            self.add_model(model)
        self.disk_hold_penalty = False
        np.random.seed() # initialize random number generator

    def init(self):
        self.add_dtype('disk_cursor_size', 'f8', (1,))
        self.add_dtype('disk_target_size', 'f8', (1,))
        super().init()

    def update_disk_cursor(self):
        
        # Remove the previous disk cursor
        for model in self.disk_cursor.graphics_models:
            self.remove_model(model)

        # Add a new cursor with updated radius
        raw_value = self.joystick.get()
        if raw_value is None:
            raw_value = 0.09
        cm_value = (raw_value + self.disk_cursor_offset) * self.disk_cursor_gain

        self.disk_cursor_radius = min(max(cm_value, self.disk_cursor_bounds[0]), self.disk_cursor_bounds[1])
        self.task_data['disk_cursor_size'] = self.disk_cursor_radius
        self.disk_cursor = VirtualCircularTarget(target_radius=self.disk_cursor_radius, target_color=disk_colors[self.disk_cursor_color], starting_pos=self.disk_pos)
        for model in self.disk_cursor.graphics_models:
            self.add_model(model)

        if self.state == "pause":
            self.disk_cursor.hide()

    def _cycle(self):
        '''
        Calls any update functions necessary and redraws screen
        '''
        self.update_disk_cursor()
        super()._cycle()


class DiskMatching(DiskMatchingMixin, Window, LogExperiment):
    """
    Annular targets are acquired by "holding" a disk cursor at the appropriate radius
    """

    status = dict(
        wait = dict(start_trial="target", start_pause="pause"),
        target = dict(enter_target="hold", start_pause="pause_init"),
        hold = dict(leave_target="hold_penalty", hold_complete="reward", start_pause="pause_init"),
        hold_penalty = dict(hold_penalty_end="wait", start_pause="pause", end_state=True),
        reward = dict(reward_end="wait", stoppable=False, end_state=True),
        pause_init = dict(continue_pause="pause"), # send a TRIAL_END event, then go immediately to pause state
        pause = dict(end_pause="wait", end_state=True),
    )

    # initial state
    state = "wait"

    hold_time = traits.Tuple((0.5, 1.5), desc="Amount of time in (min, max) seconds the target must be held before reward")
    hold_penalty_time = traits.Float(0.5, desc="Amount of time in seconds spent in hold penalty")
    reward_time = traits.Float(.5, desc="Length of reward dispensation")
   
    #### TEST FUNCTIONS ####
    def _test_start_trial(self, ts):
        return not self.pause

    def _test_enter_target(self, ts):
        d = self.disk_cursor_radius - self.disk_target_radius_trial
        return (d >= 0) and (d <= self.disk_target_tolerance)

    def _test_leave_target(self, ts):
        d = self.disk_cursor_radius - self.disk_target_radius_trial
        return (d > self.disk_target_tolerance) or (d < 0)

    def _test_hold_complete(self, ts):
        return ts > self.hold_time_trial
    
    def _test_hold_penalty_end(self, ts):
        return ts > self.hold_penalty_time

    def _test_reward_end(self, ts):
        return ts > self.reward_time
    
    def _test_start_pause(self, ts):
        return self.pause
    
    def _test_continue_pause(self, ts):
        return self.pause
    
    def _test_end_pause(self, ts):
        return not self.pause

    ### STATE FUNCTIONS ###
    def _start_wait(self):

        # Select a random parameters
        if not self.disk_hold_penalty:
            self.disk_target_radius_trial = np.random.uniform(*self.disk_target_radius)
            self.hold_time_trial = np.random.uniform(*self.hold_time)

    def _start_target(self):
        self.sync_event('TARGET_ON', 0xd)
        
        self.task_data['disk_target_size'] = self.disk_target_radius_trial
        inner_radius = self.disk_target_radius_trial
        outer_radius = self.disk_target_radius_trial + self.disk_target_tolerance
        self.disk_target = VirtualTorusTarget(inner_radius, outer_radius, target_color=disk_colors[self.disk_target_color], starting_pos=self.disk_pos)
        for model in self.disk_target.graphics_models:
            self.add_model(model)

    def _start_hold(self):
        self.sync_event('CURSOR_ENTER_TARGET', 0xd)
        self.disk_target.sphere.color = disk_colors['target_bright']
        self.disk_hold_penalty = False

    def _start_hold_penalty(self):
        self.sync_event('HOLD_PENALTY')

        # Remove the disk target
        for model in self.disk_target.graphics_models:
            self.remove_model(model)

        self.disk_hold_penalty = True

        self.task_data['disk_target_size'] = 0.

    def _end_hold_penalty(self):
        self.sync_event('TRIAL_END')

    def _start_reward(self):
        self.disk_target.cue_trial_end_success()
        self.sync_event('REWARD')
        
    def _end_reward(self):
        self.sync_event('TRIAL_END')

        # Remove the disk target
        for model in self.disk_target.graphics_models:
            self.remove_model(model)

        self.task_data['disk_target_size'] = self.disk_target_radius_trial

    def _start_pause_init(self):
        self.sync_event('TRIAL_END')

    def _start_pause(self):
        self.sync_event('PAUSE_START')

        # Remove the disk target
        for model in self.disk_target.graphics_models:
            self.remove_model(model)
        
        self.task_data['disk_target_size'] = self.disk_target_radius_trial

    def _end_pause(self):
        self.sync_event('PAUSE_END')

    @classmethod
    def get_desc(cls, params, log_summary):
        duration = round(log_summary['runtime'] / 60, 1)
        return "{}/{} succesful trials in {} min".format(
            log_summary['n_success_trials'], log_summary['n_trials'], duration)



import time
from built_in_tasks.force_task import DiskMatching
from built_in_tasks.manualcontrolmultitasks import TrackingTask, rotations, ManualControl, ScreenTargetTracking
from built_in_tasks.othertasks import Conditions, LaserConditions, SweptLaserConditions
from built_in_tasks.target_capture_task import ScreenTargetCapture
from built_in_tasks.passivetasks import YouTube
from features.generator_features import Autostart
from features.hdf_features import SaveHDF
from riglib.stereo_opengl.window import WindowDispl2D
from riglib import experiment
from features.peripheral_device_features import ForceControl, MouseControl
from features.optitrack_features import OptitrackSimulate, Optitrack, SpheresToCylinders
from features.reward_features import ProgressBar
import cProfile
import pstats
from riglib.stereo_opengl.window import Window, Window2D
import unittest
import numpy as np
import os
import socket

def init_exp(base_class, feats, seq=None, **kwargs):
    hostname = socket.gethostname()
    if hostname in ['pagaiisland2']:
        os.environ['DISPLAY'] = ':0.1'
    Exp = experiment.make(base_class, feats=feats)
    if seq is not None:
        exp = Exp(seq, **kwargs)
    else:
        exp = Exp(**kwargs)
    exp.init()
    return exp

class TestManualControlTasks(unittest.TestCase):

    @unittest.skip("")
    def test_exp(self):
        seq = ManualControl.centerout_2D()
        exp = init_exp(ManualControl, [MouseControl, Window2D], seq)
        exp.rotation = 'xzy'
        exp.run()
    
    @unittest.skip("")
    def test_tracking(self):
        print("Running tracking task test")
        seq = TrackingTask.tracking_target_debug(nblocks=1, ntrials=6, time_length=5, seed=40, sample_rate=60, ramp=1) # sample_rate needs to match fps in ScreenTargetTracking
        exp = init_exp(TrackingTask, [MouseControl], seq) # , window_size=(1000,800)
        exp.rotation = 'xzy'
        exp.run()

    @unittest.skip("")
    def test_sequence(self):
        print("Running sequence task test")
        seq = ScreenTargetCapture.sequence_2D(nblocks=1, distance=5)
        exp = init_exp(ManualControl, [MouseControl, Window2D], seq) # , window_size=(1000,800)
        exp.rotation = 'xzy'
        exp.run()

    @unittest.skip("")
    def test_force_task(self):
        print("Running force task test")
        exp = init_exp(DiskMatching, [ForceControl, Window2D, Autostart], None) # , window_size=(1000,800)
        exp.rand_start = (0.5,1)
        exp.run()
        t0 = time.time()
        while time.time() - t0 < 10:
            pass
        exp.end_task()

    @unittest.skip("only to test progress bar")
    def test_tracking(self):
        seq = TrackingTask.tracking_target_debug(nblocks=1, ntrials=6, time_length=5, seed=40, sample_rate=60, ramp=1) # sample_rate needs to match fps in ScreenTargetTracking
        exp = init_exp(TrackingTask, [MouseControl, Window2D, ProgressBar], seq)
        exp.rotation = 'xzy'
        exp.run()

    # @unittest.skip("only to test 3d window")
    def test_3d(self):
        seq = ManualControl.centerout_2D()
        exp = init_exp(ManualControl, [MouseControl, SpheresToCylinders], seq, stereo_mode='projection',
                       rotation='xyz', window_size=(1000,800), fullscreen=False, limit2d=False)
        exp.run()


class TestSeqGenerators(unittest.TestCase):

    # @unittest.skip("")
    def test_gen_ascending(self):
        seq = Conditions.gen_conditions(3, [1, 2], ascend=True)
        self.assertSequenceEqual(seq[0], [0, 0, 0, 1, 1, 1])

    # @unittest.skip("")
    def test_gen_out_2D(self):
        seq = ScreenTargetCapture.out_2D(nblocks=1, )
        seq = list(seq)
        idx = np.array([s[0][0] for s in seq])
        loc = np.array([s[1][0] for s in seq])
        print(idx)
        print(loc)
        self.assertCountEqual(idx, [1, 2, 3, 4, 5, 6, 7, 8])

        # Target 1 should be 12 o'clock
        self.assertAlmostEqual(loc[idx == 1, 0][0], 0)
        self.assertAlmostEqual(loc[idx == 1, 2][0], 10)

        # Target 3 should be 3 o'clock
        self.assertAlmostEqual(loc[idx == 3, 0][0], 10)
        self.assertAlmostEqual(loc[idx == 3, 2][0], 0)

    # @unittest.skip("")
    def test_dual_laser_wave(self):
        seq = LaserConditions.dual_laser_square_wave(duty_cycle_1=0.025, duty_cycle_2=0.025, phase_delay_2=0.1)
        print(seq[0])

    def test_swept_laser_pulse(self):
        seq = SweptLaserConditions.single_laser_pulse()
        print(seq[0])

    def test_corners(self):
        seq = ScreenTargetCapture.corners_2D(chain_length=3)
        seq = list(seq)

        idx = np.array([s[0][0] for s in seq])
        loc = np.array([s[1][0] for s in seq])
        print("corners---------------")
        print(idx)
        print(loc)
        print("---------------corners")

class TestYouTube(unittest.TestCase):

    @unittest.skip("")
    def test_youtube_exp(self):

        exp = init_exp(YouTube, [], youtube_url="https://www.youtube.com/watch?v=Qe9ansjvF7M")
        exp.run()

if __name__ == '__main__':
    unittest.main()


'''
Graphical display classes. Experimental tasks involving graphical displays
inherit from these classes.
'''
import os

import numpy as np
from OpenGL.GL import *

from ..experiment import LogExperiment
from ..experiment import traits

from .render import stereo, render
from .models import Group
from .xfm import Quaternion
from .primitives import Sphere, Cube, Chain
from .environment import Box
from .primitives import Cylinder, Sphere, Cone
import socket

try:
    os.environ['PYGAME_HIDE_SUPPORT_PROMPT'] = "hide"
    import pygame
except ImportError:
    import warnings
    warnings.warn('riglib/stereo_opengl/window.py: not importing name pygame')

from config.rig_defaults import window as defaults

# for WindowDispl2D only
from .primitives import Shape2D

class Window(LogExperiment):
    '''
    Generic stereo window
    '''
    status = dict(draw=dict(stop=None))
    state = "draw"
    stop = False

    # XPS computer
    # window_size = (1920*2, 1080)
    window_size = traits.Tuple(defaults['window_size'], desc='Window size, in pixels')
    background = traits.Tuple((0.,0.,0.,1.), desc="Background color (R,G,B,A)")
    fullscreen = traits.Bool(True, desc="Fullscreen window")

    #Screen parameters, all in centimeters -- adjust for monkey
    screen_dist = traits.Float(defaults['screen_dist'], desc="Screen to eye distance (cm)")
    screen_half_height = traits.Float(defaults['screen_half_height'], desc="Screen half height (cm)")
    iod = traits.Float(2.5, desc="Intraocular distance (cm)")     # intraocular distance
    stereo_mode = traits.OptionsList(['mirror', 'projection', 'anaglyph'], desc="Stereo mode", 
                                     bmi3d_input_options=['mirror', 'projection', 'anaglyph'])

    show_environment = traits.Int(0, desc="Show wireframe box around environment")

    hidden_traits = ['stereo_mode', 'screen_dist', 'screen_half_height', 'iod', 'show_environment', 'background']

    def __init__(self, *args, **kwargs):
        self.display_start_pos = kwargs.pop('display_start_pos', "0,0")
        super(Window, self).__init__(*args, **kwargs)

        self.models = []
        self.world = None
        self.event = None

        # os.popen('sudo vbetool dpms on')
        self.fov = np.degrees(np.arctan(self.screen_half_height/self.screen_dist))*2
        self.screen_cm = [2 * self.screen_half_height * self.window_size[0]/self.window_size[1], 2 * self.screen_half_height]

        if self.show_environment:
            self.add_model(Box())

    def set_os_params(self):
        os.environ['SDL_VIDEO_WINDOW_POS'] = self.display_start_pos
        os.environ['SDL_VIDEO_X11_WMCLASS'] = "monkey_experiment"

    def screen_init(self):
        self.set_os_params()
        pygame.init()

        pygame.display.gl_set_attribute(pygame.GL_DEPTH_SIZE, 24)
        flags = pygame.DOUBLEBUF | pygame.HWSURFACE | pygame.OPENGL | pygame.NOFRAME
        if self.fullscreen:
            flags = flags | pygame.FULLSCREEN
        try:
            pygame.display.gl_set_attribute(pygame.GL_MULTISAMPLEBUFFERS,1)
            self.screen = pygame.display.set_mode(self.window_size, flags)
        except:
            pygame.display.gl_set_attribute(pygame.GL_MULTISAMPLEBUFFERS,0)
            self.screen = pygame.display.set_mode(self.window_size, flags)

        glEnable(GL_BLEND)
        glDepthFunc(GL_LESS)
        glEnable(GL_DEPTH_TEST)
        glEnable(GL_TEXTURE_2D)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
        glClearColor(*self.background)
        glClearDepth(1.0)
        glDepthMask(GL_TRUE)
        glEnable(GL_CULL_FACE) # temporary solution to alpha blending issue with spheres. just draw the front half of the sphere
        glCullFace(GL_BACK)

        self.renderer = self._get_renderer()

        #this effectively determines the modelview matrix
        self.world = Group(self.models)
        self.world.init()

        #up vector is always (0,0,1), why would I ever need to roll the camera?!
        self.set_eye((0, -self.screen_dist, 0), (0,0))

        pygame.mouse.set_visible(False)

    def _get_renderer(self):
        near = 1
        far = 1024
        if self.stereo_mode == 'mirror':
            glCullFace(GL_FRONT)
            return stereo.MirrorDisplay(self.window_size, self.fov, near, far, self.screen_dist, self.iod)
        elif self.stereo_mode == 'projection':
            return render.Renderer(self.window_size, self.fov, near, far)
        elif self.stereo_mode == 'anaglyph':
            return stereo.Anaglyph(self.window_size, self.fov, near, far, self.screen_dist, self.iod)
        else:
            raise ValueError("Unknown stereo mode: %s" % self.stereo_mode)

    def set_eye(self, pos, vec, reset=True):
        '''Set the eye's position and direction. Camera starts at (0,0,0), pointing towards positive y'''
        self.world.translate(pos[0], pos[2], pos[1], reset=True).rotate_x(-90)
        self.world.rotate_y(vec[0]).rotate_x(vec[1])

    def add_model(self, model):
        if self.world is None:
            #world doesn't exist yet, add the model to cache
            self.models.append(model)
        else:
            #We're already running, initialize the model and add it to the world
            model.init()
            self.world.add(model)

    def remove_model(self, model):
        '''remove the second model from the last'''
        self.world.remove(model)

    def show_object(self, obj, show=False):
        '''
        Show or hide an object. This function is an abstraction so that tasks don't need to know about attach/detach
        '''
        if show:
            obj.attach()
        else:
            obj.detach()

    def draw_world(self):
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
        self.renderer.draw(self.world)
        pygame.display.flip()
        self.renderer.draw_done()

    def _get_event(self):
        for e in pygame.event.get(pygame.KEYDOWN):
            return (e.key, e.type)

    def _start_None(self):
        pygame.display.quit()

    def _test_stop(self, ts):
        '''
        Stop the task if the escape key is pressed, or if the super _test_stop instructs a stop
        '''
        super_stop = super(Window, self)._test_stop(ts)
        from pygame import K_ESCAPE
        key_stop = self.event is not None and self.event[0] == K_ESCAPE
        if key_stop:
            print("Window closed. Stopping task")
        return super_stop or key_stop
    
    def update_report_stats(self):
        self.reportstats['FPS'] = round(self.clock.get_fps(),2)
        return super().update_report_stats()
    
    def requeue(self):
        self.renderer._queue_render(self.world)

    def _cycle(self):
        super(Window, self)._cycle() # is this order intentional? why not cycle first then draw the screen?
        self.event = self._get_event()
        self.requeue()
        self.draw_world()


class WindowWithExperimenterDisplay(Window):
    hostname = socket.gethostname()
    # This class has a hard-coded window size
    if hostname == 'lynx':
        _stereo_window_flip = True
        _stereo_main_flip_z = True
        window_size = (1280+480, 1024)
        window_size1 = (1280, 1024)
        window_size2 = (480, 270)
    else:
        _stereo_window_flip = False
        _stereo_main_flip_z = False
        window_size = (1920 + 480, 1080)
        window_size1 = (1920, 1080)
        window_size2 = (480, 270)

    def __init__(self, *args, **kwargs):
        super(WindowWithExperimenterDisplay, self).__init__(*args, **kwargs)
        # This class has a hard-coded window size
        # self.window_size = (1920 + 480, 1080)

    def set_os_params(self):
        # NOTE: in Ubuntu Unity, setting the SDL_VIDEO_WINDOW_POS seems to be largely ignored.
        # You can set which screen the window appears on if you have a dual display, but you cannot set the exact position
        # Instead, you have to hard-code a render start location in the compiz-config settings manager
        # http://askubuntu.com/questions/452995/how-to-adjust-window-placement-in-unity-ubuntu-14-04-based-on-overlapping-top-b
        os.environ['SDL_VIDEO_WINDOW_POS'] = self.display_start_pos
        os.environ['SDL_VIDEO_X11_WMCLASS'] = "monkey_experiment_with_mini"

    def _get_renderer(self):
        near = 1
        far = 1024
        return stereo.DualMultisizeDisplay(self.window_size1, self.window_size2, self.fov, near, far, self.screen_dist, self.iod, flip=self._stereo_window_flip,
            flip_main_z = self._stereo_main_flip_z)


TRANSPARENT = (106,0,42)

class WindowDispl2D():
    '''For testing only -- draws world on a 2D screen. May cause mild confusion -- transforms 
    incoming 3D coordinates (x,y,z) into 2D coordinates (x,y) by mapping z onto y'''

    def __init__(self, *args, **kwargs):
        self.is_pygame_display = True
        super().__init__(*args, **kwargs)

    def screen_init(self):
        os.environ['SDL_VIDEO_WINDOW_POS'] = self.display_start_pos
        os.environ['SDL_VIDEO_X11_WMCLASS'] = "monkey_experiment"
        pygame.init()
        self.clock = pygame.time.Clock()

        flags = pygame.NOFRAME | pygame.DOUBLEBUF | pygame.HWSURFACE

        if self.fullscreen:
            flags = flags | pygame.FULLSCREEN

        self.display_border = 10

        self.size = np.array(self.window_size, dtype=np.float64)
        self.screen = pygame.display.set_mode(self.window_size, flags)
        self.screen_background = pygame.Surface(self.screen.get_size()).convert()
        self.screen_background.fill(self.background)

        x_max, y_max = (self.screen_cm[0]/2, self.screen_cm[1]/2)
        x_min, y_min = (-self.screen_cm[0]/2, -self.screen_cm[1]/2)
        self.normalize = np.array(np.diag([1./(x_max-x_min), 1./(y_max-y_min), 1]))
        self.center_xform = np.array([[1., 0, -x_min], 
                                      [0, 1., -y_min],
                                      [0, 0, 1]])
        self.norm_to_screen = np.array(np.diag(np.hstack([self.size, 1])))

        # the y-coordinate in pixel space has to be swapped graphics convention reasons
        self.flip_y_coord = np.array([[1, 0, 0],
                                      [0, -1, self.size[1]],
                                      [0, 0, 1]])
        self.pos_space_to_pixel_space = np.dot(self.flip_y_coord, np.dot(self.norm_to_screen, np.dot(self.normalize, self.center_xform)))

        self.world = Group(self.models)
        # Dont 'init' self.world in this Window. Just allocates a bunch of OpenGL stuff which is not necessary (and may not work in some cases)
        # self.world.init()

        #initialize surfaces for translucent markers
        self.surf={}
        self.surf['0'] = pygame.Surface(self.screen.get_size())
        self.surf['0'].fill(TRANSPARENT)
        self.surf['0'].set_colorkey(TRANSPARENT)

        self.surf['1'] = pygame.Surface(self.screen.get_size())
        self.surf['1'].fill(TRANSPARENT)
        self.surf['1'].set_colorkey(TRANSPARENT)

         #values of alpha: higher = less translucent
        self.surf['0'].set_alpha(170) #Cursor
        self.surf['1'].set_alpha(130) #Targets

        self.surf_background = pygame.Surface(self.surf['0'].get_size()).convert()
        self.surf_background.fill(TRANSPARENT)

        pygame.mouse.set_visible(False)

        self.i = 0

    def pos2pix(self, kfpos):
        # re-specify the point in homogenous coordinates
        pt = np.hstack([kfpos, 1]).reshape(-1, 1)

        # perform the homogenous transformation
        pix_coords = np.dot(self.pos_space_to_pixel_space, pt)

        pix_pos = np.array(pix_coords[:2,0], dtype=int)
        return pix_pos

    def get_surf(self):
        return self.surf[str(np.min([self.i,1]))]

    def draw_model(self, model):
        '''
        Draw a single Model on the current surface, or recurse if the model is a composite model (i.e., a Group)
        '''
        color = tuple([int(255*x) for x in model.color[0:3]])
        if isinstance(model, Sphere):
            pos = model._xfm.move[[0,2]]
            pix_pos = self.pos2pix(pos)

            rad = model.radius
            pix_radius = self.pos2pix(np.array([model.radius, 0]))[0] - self.pos2pix([0,0])[0]

            #Draws cursor and targets on transparent surfaces
            pygame.draw.circle(self.get_surf(), color, pix_pos, pix_radius)

        elif isinstance(model, Shape2D):
            # model.draw() returns True if the object was drawn
            #   (which happens if the object's .visible attr is True)
            if model.draw(self.get_surf(), self.pos2pix):
                pass
        elif isinstance(model, Cube):
            pos = model.xfm.move[[0,2]]
            side_len = model.side_len

            left = pos[0] - side_len/2
            right = pos[0] + side_len/2
            top = pos[1] + side_len/2
            bottom = pos[1] - side_len/2

            top_left = np.array([left, top])
            bottom_right = np.array([right, bottom])
            top_left_pix_pos = self.pos2pix(top_left)
            bottom_right_pix_pos = self.pos2pix(bottom_right)

            rect = pygame.Rect(top_left_pix_pos, bottom_right_pix_pos - top_left_pix_pos)
            color = tuple([int(255*x) for x in model.color[0:3]])

            pygame.draw.rect(self.get_surf(), color, rect)

        elif isinstance(model, (Cylinder, Cone)):
            vec_st = np.array([0., 0, 0, 1])
            vec_end = np.array([0., 0, model.height, 1])

            cyl_xform = model._xfm.to_mat()
            cyl_start = np.dot(cyl_xform, vec_st)
            cyl_end = np.dot(cyl_xform, vec_end)
            pix_radius = self.pos2pix(np.array([model.radius, 0]))[0] - self.pos2pix([0,0])[0]

            pygame.draw.line(self.get_surf(), color, self.pos2pix(cyl_start[[0,2]]), self.pos2pix(cyl_end[[0,2]]), pix_radius)

            # print cyl_start, cyl_end

        elif isinstance(model, Group):
            for mdl in model:
                self.draw_model(mdl)

    def _draw_other(self):
        pass

    def draw_world(self):
        #Refreshes the screen with original background
        self.screen.blit(self.screen_background, (0, 0))
        self.surf['0'].blit(self.surf_background,(0,0))
        self.surf['1'].blit(self.surf_background,(0,0))

        # surface index
        self.i = 0

        for model in self.world.models:
            self.draw_model(model)
            self.i += 1

        #Renders the new surfaces
        self.screen.blit(self.surf['0'], (0,0))
        self.screen.blit(self.surf['1'], (0,0))

        self._draw_other()
        pygame.display.update()

    def requeue(self):
        '''
        Simulation 'requeue' does nothing because the simulation is lazy and
        inefficient and chooses to redraw the entire screen every loop
        '''
        pass


class Window2D():
    '''
    Render the world in 2D without lighting
    '''
    def _get_renderer(self):
        glCullFace(GL_BACK)
        return render.Renderer2D(self.screen_cm)
    
class FakeWindow(Window):
    '''
    A dummy class to secretly avoid rendering graphics without
    the graphics-based tasks knowing about it. Used e.g. for simulation
    purposes where the graphics only slow down the simulation.
    '''
    background = (1,1,1,1)
    def __init__(self, *args, **kwargs):
        self.models = []
        self.world = None
        self.event = None
        super(FakeWindow, self).__init__(*args, **kwargs)

    def screen_init(self):
        self.world = Group(self.models)
        # self.world.init()

    def draw_world(self):
        pass

    def requeue(self):
        pass

    def _start_reward(self, *args, **kwargs):
        n_rewards = self.calc_state_occurrences('reward')
        if n_rewards % 10 == 0:
            print(n_rewards)
        super(FakeWindow, self)._start_reward(*args, **kwargs)

    def show_object(self, *args, **kwargs):
        pass

    def _get_event(self):
        pass


class FPScontrol(Window):
    '''A mixin that adds a WASD + Mouse controller to the window.
    Use WASD to move in XY plane, q to go down, e to go up'''

    def init(self):
        super(FPScontrol, self).init()
        pygame.mouse.set_visible(False)
        self.eyepos = [0,-self.screen_dist, 0]
        self.eyevec = [0,0]
        self.wasd = [False, False, False, False, False, False]

    def _get_event(self):
        retme = None
        for e in pygame.event.get([pygame.MOUSEMOTION, pygame.KEYDOWN, pygame.KEYUP, pygame.QUIT]):
            moved = any(self.wasd)
            if e.type == pygame.MOUSEMOTION:
                self.world.xfm.rotate *= Quaternion.from_axisangle((1,0,0), np.radians(e.rel[1]*.1))
                self.world.xfm.rotate *= Quaternion.from_axisangle((0,0,1), np.radians(e.rel[0]*.1))
                self.world._recache_xfm()
            elif e.type == pygame.KEYDOWN:
                kn = pygame.key.name(e.key)
                if kn in ["escape", "q"]:
                    self.stop = True
                retme = (e.key, e.type)
            elif e.type == pygame.QUIT:
                self.stop = True

            if moved:
                self.set_eye(self.eyepos, self.eyevec, reset=True)
        return retme



# Generated by Django 3.1.4 on 2020-12-12 18:48

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('tracker', '0006_datafile_backup_status'),
    ]

    operations = [
        migrations.AddField(
            model_name='taskentry',
            name='template',
            field=models.BooleanField(blank=True, default=False),
        ),
    ]



'''
Generate reporting stats for completed/ongoing tasks
'''


import numpy as np
from .experiment import LogExperiment, TrialTypes
import re

def trialtype(exp):
    '''
    Docstring

    Parameters
    ----------

    Returns
    -------
    '''
    assert isinstance(exp, LogExperiment), "Cannot report on non-logged experiments"
    ttypes = exp.trial_types if isinstance(exp, TrialTypes) else ["trial"]
    trials = dict([(t, dict(correct=[], incorrect=[], timeout=[])) for t in ttypes])
    report = dict(rewards=0, prematures=[], trials=trials)
    trial = None

    for state, event, t in exp.event_log:
        if trial is not None:
            if "incorrect" in event:
                report['trials'][state]["incorrect"].append(t - trial)
            elif "correct" in event:
                report['trials'][state]["correct"].append(t - trial)
            elif "timeout" in event:
                report['trials'][state]["timeout"].append(t - trial)
            trial = None
        
        if event == "start_trial":
            trial = t
        elif state == "reward":
            report['rewards'] += 1
    
    return report

def general(Exp, event_log, repdict, ntrials, nrewards, reward_len):
    '''
    Docstring

    Parameters
    ----------

    Returns
    -------
    '''
    report = dict(trials = 0, reward_len=[])

    report['trials'] = ntrials
    report['reward_len'] = reward_len, nrewards

    if report['trials'] == 0:
        report['rates'] = (0,0)
    else:
       report['rates'] = (float(nrewards)/ntrials, 1-(float(nrewards)/ntrials))

    return report

def print_report(report):
    '''Prints a report generated by report(exp)

    Docstring

    Parameters
    ----------

    Returns
    -------
    '''
    repstr = ["%8s: %d"%("rewards", report['rewards'])]
    ttrial = 0
    for tname, tdict in list(report['trials'].items()):
        total = len(tdict['correct']) + len(tdict['incorrect']) + len(tdict['timeout'])
        ttrial += total
        if total == 0:
            cper = 0
        else:
            cper = float(len(tdict['correct'])) / total * 100
        cRT = np.mean(tdict['correct'])
        repstr.append("%8s: %g%%, RT=%g"%(tname, cper, cRT))
    repstr.insert(0, "%8s: %d"%("total", ttrial))
    print("\n".join(repstr))



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="fsm",
    version="0.1.0",
    author="Carmena Lab",
    description="Python Finite State Machine",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
)



'''
Code for interacting with the Phdigets API
'''
import time
import numpy as np

from Phidgets.Devices.InterfaceKit import InterfaceKit
from .source import DataSourceSystem

class System(DataSourceSystem):
    '''
    Generic DataSourceSystem interface for the Phidgets board: http://www.phidgets.com/products.php?category=0&product_id=1018_2
    '''
    update_freq = 1000

    def __init__(self, n_sensors=2, n_inputs=1):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        self.n_sensors = n_sensors
        self.n_inputs = n_inputs
        self.interval = 1. / self.update_freq

        self.sensordat = np.zeros((n_sensors,))
        self.inputdat = np.zeros((n_inputs,), dtype=np.bool)
        self.data = np.zeros((1,), dtype=self.dtype)

        self.kit = InterfaceKit()
        self.kit.openPhidget()
        self.kit.waitForAttach(2000)
    
    def start(self):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        self.tic = time.time()

    def stop(self):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        pass
    
    def get(self):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        toc = time.time() - self.tic
        if 0 < toc < self.interval:
            time.sleep(self.interval - toc)
        try:
            for i in range(self.n_sensors):
                self.sensordat[i] = self.kit.getSensorValue(i) / 1000.
            for i in range(self.n_inputs):
                self.inputdat[i] = self.kit.getInputState(i)
        except:
            print('sensor_error')
        self.data['sensors'] = self.sensordat
        self.data['inputs'] = self.inputdat
        self.tic = time.time()
        return self.data
    
    def sendMsg(self, msg):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        pass

    def __del__(self):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        self.kit.closePhidget()

def make(sensors, inputs, cls=System, **kwargs):
    '''
    Docstring
    This ridiculous function dynamically creates a class with a new init function

    Parameters
    ----------

    Returns
    -------
    '''
    def init(self, **kwargs):
        super(self.__class__, self).__init__(n_sensors=sensors, n_inputs=inputs, **kwargs)
    
    dtype = np.dtype([('sensors', np.float, (sensors,)), ('inputs', np.bool, (inputs,))])
    return type(cls.__name__, (cls,), dict(dtype=dtype, __init__=init))



import numpy as np
import sys, time
import socket


N_TEST_FRAMES = 1 #number of testing frames during start
class System(object):
    """
    this is is the dataSource interface for getting the mocap at BMI3D's reqeust
    compatible with DataSourceSystem
    uses data_array to keep track of the lastest buffer
    """
    port_num  = 1230 #same as the optitrack #default to 1230
    HEADERSIZE = 10
    rece_byte_size = 512
    debug = True
    optitrack_ip_addr = "10.155.206.1"
    TIME_OUT_TIME = 2

    #a list of supported commands
    SUPPORTED_COMMANDS = [
        "start_rec",
        "send_markers",
        'send_rigid_bodies',
        'stop',
        'start'
    ]

    SUPPORTED_STREAM_TYPES = [
        'rigid_bodies',
        'markers'
        #
    ]



    rigidBodyCount = 1
    update_freq = 120 
    dtype = np.dtype((np.float, (rigidBodyCount, 6))) #6 degress of freedom
    
    def __init__(self):
        self.rigid_body_count = 1 #for now,only one rigid body


    def start(self, stream_type = "rb"):
        """
        stream_type
        """
        #start to connect to the client
                #set up the socket
        self.s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.s.settimeout(self.TIME_OUT_TIME)
  
        print("connecting to the c# server")
        '''
        self.s.bind(('', 1230)) #bind to all incoming request
        self.s.listen() #listen to one clinet
        '''
        try:
            #clientsocket, address = self.s.accept()
            self.s.connect((self.optitrack_ip_addr, self.port_num))
        except:
            print("cannot connect to Motive")
            print("Is the c# server running?")
        
        #otherwise it works as expected and set client to be a 
        #class property
        #self.clientsocket = clientsocket
        print(f"Connection to c# client \
             {self.optitrack_ip_addr} has been established.")


        if stream_type in self.SUPPORTED_STREAM_TYPES: 
            self.send_command('send_'+stream_type)
        else:
            raise Exception(f'{stream_type} is not supported \n\n suported stream types are\n{self.SUPPORTED_STREAM_TYPES}')

            

        #automatically pull 10 frames
        # and cal the mean round trip time
        t1 = time.perf_counter()
        for i in range(N_TEST_FRAMES): self.get()
        t2 = time.perf_counter()
        print(f'time to grab {N_TEST_FRAMES} frames : \
              {(t2 - t1)} s ')
                        

    def stop(self):
        msg = "stop"
        self.send_command(msg)
        #close the socket
        #self.s.close()
        print("socket closed!")
    
    def get(self):
        #the property  that gets one frame of data
        # 3 positions and 3 angles
        #the last element is frame number
        msg = "get"
        result_string = self.send_and_receive(msg)
        motive_frame = np.fromstring(result_string, sep=',')
        current_value = motive_frame[:6] #only using the motion data
        current_value.transpose()


        #for some weird reason, the string needs to be expanded..
        #just send the motion data for now
        current_value = np.expand_dims(current_value, axis = 0)
        current_value = np.expand_dims(current_value, axis = 0)
        return current_value
        


    def send_command(self, msg):
        """
        a function that sends the command to remote optitrack
        msg(string): of the message
                    the msg string has to be in the self.SUPPORTED_COMMANDS
                    
        """
        #check if the command is supported
        if msg not in self.SUPPORTED_COMMANDS: raise Exception(f'{msg} not supported')


        #get the message in string and encode in  bytes and send to the socket
        msg = f"{len(msg):<{self.HEADERSIZE}}"+msg
        msg_ascii = msg.encode("ascii")
        self.s.send(msg_ascii)

    def send_and_receive(self, msg):
        #this function sends a command
        #and then wait for a response
        msg = f"{len(msg):<{self.HEADERSIZE}}"+msg
        msg_ascii = msg.encode("ascii")
        self.s.send(msg_ascii)
        result_in_bytes = self.s.recv(self.rece_byte_size)
        return str(result_in_bytes,encoding="ASCII")
    

class Simulation(System):
    '''
    this class does all the things except when the optitrack is not broadcasting data
    the get function starts to return random numbers
    '''
    update_freq = 60 #Hz
    def get(self):
        mag_fac = 10
        current_value = np.random.rand(self.rigidBodyCount, 6) * mag_fac
        current_value = np.expand_dims(current_value, axis = 0)
        return current_value
        

if __name__ == "__main__":
    s = System()
    s.start()
    s.send_command("start_rec")
    #s.send_command("send_markers")
    time.sleep(5)
    s.stop()
    print("finished")



'''
Feedback controllers. In the BMI context, these have three potential
applications: (1) assisitve/shared control between neural and pure machine sources,
(2) estimating the "intended" BMI movements of the subject (see clda.py), and
(3) driving populations of simulated neurons with artificially constructed tuning
functions. 
'''
import numpy as np


class FeedbackController(object):
    '''
    Abstract class for feedback controllers, only used for type-checking & interface standardization
    '''
    def __init__(self, *args, **kwargs):
        pass

    def calc_next_state(self, current_state, target_state, mode=None):
        raise NotImplementedError

    def __call__(self, current_state, target_state, mode=None):
        raise NotImplementedError

    def get(self, current_state, target_state, mode=None):
        raise NotImplementedError


class LinearFeedbackController(FeedbackController):
    '''
    Generic linear state-feedback controller. Can be time-varying in general and not be related to a specific cost function
    '''
    def __init__(self, A, B, F):
        '''
        Constructor for LinearFeedbackController
        
        FC for a linear system
            x_{t+1} = Ax_t + Bu_t
        where the control input u_t is calculated using linear feedback
            u_t = F(x_t - x^*)

        Parameters
        ----------
        B : np.mat
            Input matrix for the system
        F : np.mat
            static feedback gain matrix

        Returns
        -------
        LinearFeedbackController instance
        '''
        self.A = A
        self.B = B
        self.F = F 

    def calc_next_state(self, current_state, target_state, mode=None):
        '''
        Returns x_{t+1} = Ax_t + BF(x* - x_t)

        Parameters
        ----------
        current_state : np.matrix
            Current state of the system, x_t 
        target_state : np.matrix
            State that you're trying to steer the system toward, x*
        mode : object, default=None
            Select the operational mode of the feedback controller. Ignored in this class

        Returns
        -------
        np.matrix
        '''
        # explicitly cast current_state and target_state to column vectors
        current_state = np.mat(current_state).reshape(-1,1)
        target_state = np.mat(target_state).reshape(-1,1)
        ns = self.A * current_state + self.B * self.F * (target_state - current_state)

        return ns

    def __call__(self, current_state, target_state, mode=None):
        '''
        Calculate the control input u_t = BF(x* - x_t)

        Parameters
        ----------
        (see self.calc_next_state for input argument descriptions)

        Returns
        -------
        np.mat of shape (N, 1)
            B*u where u_t = F(x^* - x_t)
        '''
        
        
        current_state = np.mat(current_state).reshape(-1,1)
        target_state = np.mat(target_state).reshape(-1,1)
        Bu = self.B * self.F * (target_state - current_state)
        return Bu


class LQRController(LinearFeedbackController):
    '''Linear feedback controller where control gains are set by optimizing a quadratic cost function'''
    def __init__(self, A, B, Q, R, **kwargs):
        '''
        Constructor for LQRController

        The system should evolve as
        $$x_{t+1} = Ax_t + Bu_t + w_t; w_t ~ N(0, W)$$

        with infinite horizon cost 
        $$\sum{t=0}^{+\infty} (x_t - x_target)^T * Q * (x_t - x_target) + u_t^T * R * u_t$$

        Parameters
        ----------
        A: np.ndarray of shape (n_states, n_states)
            Model of the state transition matrix of the system to be controlled. 
        B: np.ndarray of shape (n_states, n_controls)
            Control input matrix of the system to be controlled. 
        Q: np.ndarray of shape (n_states, n_states)
            Quadratic cost on state
        R: np.ndarray of shape (n_controls, n_controls)
            Quadratic cost on control inputs

        Returns
        -------
        LQRController instance
        '''
        self.A = np.mat(A)
        self.B = np.mat(B)
        self.Q = np.mat(Q)
        self.R = np.mat(R)
        F = self.dlqr(A, B, Q, R, **kwargs)
        super(LQRController, self).__init__(A, B, F, **kwargs)

    @staticmethod
    def dlqr(A, B, Q, R, Q_f=None, T=np.inf, max_iter=1000, eps=1e-10, dtype=np.mat):
        '''
        Find the solution to the discrete-time LQR problem

        The system should evolve as
        $$x_{t+1} = Ax_t + Bu_t + w_t; w_t ~ N(0, W)$$

        with cost function
        $$\sum{t=0}^{T} (x_t - x_target)^T * Q * (x_t - x_target) + u_t^T * R * u_t$$

        The cost function can be either finite or infinite horizion, where finite horizion is assumed if 
        a final const is specified

        Parameters
        ----------
        A: np.ndarray of shape (n_states, n_states)
            Model of the state transition matrix of the system to be controlled. 
        B: np.ndarray of shape (n_states, n_controls)
            Control input matrix of the system to be controlled. 
        Q: np.ndarray of shape (n_states, n_states)
            Quadratic cost on state
        R: np.ndarray of shape (n_controls, n_controls)
            Quadratic cost on control inputs
        Q_f: np.ndarray of shape (n_states, n_states), optional, default=None
            Final quadratic cost on state at the end of the horizon. Only applies to finite-horizion variants
        T: int, optional, default = np.inf
            Control horizon duration. Infinite by default. Must be less than infinity (and Q_f must be specified)
            to get the finite horizon feedback controllers
        eps: float, optional, default=1e-10
            Threshold of change in feedback matrices to define when the Riccatti recursion has converged
        dtype: callable, optional, default=np.mat
            Callable function to reformat the feedback matrices 

        Returns
        -------
        K: list or matrix
            Returns a sequence of feedback gains if finite horizon or a single controller if infinite horizon.

        '''
        if Q_f == None: 
            Q_f = Q

        if T < np.inf: # Finite horizon
            K = [None]*T
            P = Q_f
            for t in range(0,T-1)[::-1]:
                K[t] = (R + B.T*P*B).I * B.T*P*A
                P = Q + A.T*P*A -A.T*P*B*K[t]
            return dtype(K)
        else: # Infinite horizon
            P = Q_f
            K = np.inf
            for t in range(max_iter):
                K_old = K
                K = (R + B.T*P*B).I * B.T*P*A
                P = Q + A.T*P*A -A.T*P*B*K 
                if np.linalg.norm(K - K_old) < eps:
                    break
            return dtype(K)


class MultiModalLFC(LinearFeedbackController):
    '''
    A linear feedback controller with different feedback gains in different "modes"
    '''
    def __init__(self, A=None, B=None, F_dict=dict()):
        '''
        Constructor for MultiModalLFC

        Parameters
        ----------
        B : np.mat
            Input matrix for the system
        F : dict
            keys should be control 'modes', values should be np.mat static feedback gain matrices

        Returns
        -------
        MultiModalLFC instance
        '''
        self.A = A
        self.B = B
        self.F_dict = F_dict
        self.F = None

    def calc_next_state(self, current_state, target_state, mode=None):
        '''
        See LinearFeedbackController.calc_next_state for docs
        '''
        self.F = self.F_dict[mode]
        super(MultiModalLFC, self).calc_next_state(current_state, target_state, mode=mode)

    def __call__(self, current_state, target_state, mode=None):
        '''
        See LinearFeedbackController.__call__ for docs
        '''        
        self.F = self.F_dict[mode]
        super(MultiModalLFC, self).__call__(current_state, target_state, mode=mode)        


class PIDController(FeedbackController):
    '''
    Linear feedback controller where gains are set directly instead of through a cost function
    '''
    def __init__(self, K_prop, K_deriv, K_int, state_order):
        '''
        Constructor for PIDController

        Parameters
        ----------
        K_prop : float 
            Gain on proportional error
        K_deriv : float 
            Gain on derivative error 
        K_int : float
            Gain on integrated error
        state_order : np.ndarray of shape (N, 1)
            Specify whether each element of the state vector is a proportional, derivative, or integral state

        Returns
        -------
        PIDController instance
        '''
        self.K_prop = K_prop
        self.K_deriv = K_deriv
        self.K_int = K_int

        self.int_terms, = np.nonzero(state_order == -1)
        self.prop_terms, = np.nonzero(state_order == 0)
        self.deriv_terms, = np.nonzero(state_order == 1)

    def __call__(self, current_state, target_state):
        '''
        Determine the PID controller output to be added onto the current state

        Parameters
        ----------
        current_state : np.matrix
            Current state of the system, x_t 
        target_state : np.matrix
            State that you're trying to steer the system toward, x*

        Returns
        -------
        cmd : np.ndarray of shape (K, 1)
            K is the number of states in the proportional term. 
        '''        
        state_diff = target_state - current_state
        cmd = 0
        if len(self.prop_terms) > 0:
            cmd += self.K_prop * state_diff[self.prop_terms] 

        if len(self.deriv_terms) > 0:
            cmd += self.K_deriv * state_diff[self.deriv_terms] 

        if len(self.int_terms) > 0:
            cmd += self.K_int * state_diff[self.int_terms]

        return cmd

    def calc_next_state(self, current_state, target_state, **kwargs):
        '''
        see self.__call__
        '''
        return self.__call__(current_state, target_state)


######################################################################
##### Deprecated task/plant-specific controllers for simulations #####
######################################################################
class CenterOutCursorGoal(object):
    '''
    Cursor controller which moves the cursor toward the target at a constant speed
    '''
    def __init__(self, angular_noise_var=0, gain=0.15):
        '''
        Constructor for CenterOutCursorGoal

        Parameters
        ----------
        angular_noise_var: float, optional, default=0
            Angular noise is added onto the control direction as a clipped Gaussian distribution with this variance
        gain: float, optional, default=0.15
            Speed at which to move the cursor, in m/s

        Returns
        -------
        CenterOutCursorGoal instance
        '''
        self.angular_noise_var = angular_noise_var
        self.gain = gain

    def get(self, cur_target, cur_pos, keys_pressed=None):
        # Make sure y-dimension is 0
        assert cur_pos[1] == 0
        assert cur_target[1] == 0

        dir_to_targ = cur_target - cur_pos

        if self.angular_noise_var > 0:
            angular_noise_rad = np.random.normal(0, self.angular_noise_var)
            while abs(angular_noise_rad) > np.pi:
                angular_noise_rad = np.random.normal(0, self.angular_noise_var)
        else:
            angular_noise_rad = 0
        angle = np.arctan2(dir_to_targ[2], dir_to_targ[0])
        sum_angle = angle + angular_noise_rad
        return self.gain*np.array([np.cos(sum_angle), np.sin(sum_angle)])


class CenterOutCursorGoalJointSpace2D(CenterOutCursorGoal):
    '''2-link arm controller which moves the endpoint toward a target position at a constant speed'''
    def __init__(self, link_lengths, shoulder_anchor, *args, **kwargs):
        '''
        Constructor for CenterOutCursorGoalJointSpace2D

        Parameters
        ----------
        link_lengths: 
        shoulder_anchor: 
        args, kwargs: positional and keyword arguments for parent constructor (CenterOutCursorGoal)


        Returns
        -------
        '''
        self.link_lengths = link_lengths
        self.shoulder_anchor = shoulder_anchor
        super(CenterOutCursorGoalJointSpace2D, self).__init__(*args, **kwargs)


    def get(self, cur_target, cur_pos, keys_pressed=None):
        '''
        cur_target and cur_pos should be specified in workspace coordinates
        '''
        vx, vz = super(CenterOutCursorGoalJointSpace2D, self).get(cur_target, cur_pos, keys_pressed)
        vy = 0

        px, py, pz = cur_pos

        pos = np.array([px, py, pz]) - self.shoulder_anchor
        vel = np.array([vx, vy, vz])

        # Convert to joint velocities
        from riglib.stereo_opengl import ik
        joint_pos, joint_vel = ik.inv_kin_2D(pos, self.link_lengths[0], self.link_lengths[1], vel)
        return joint_vel[0]['sh_vabd'], joint_vel[0]['el_vflex']

class PosFeedbackController(FeedbackController):
    '''
    Dumb controller that just spits back the target
    '''
    def __init__(self, *args, **kwargs):
        pass

    def calc_next_state(self, current_state, target_state, mode=None):
        return target_state

    def __call__(self, current_state, target_state, mode=None):
        return target_state

    def get(self, current_state, target_state, mode=None):
        return target_state


# boot_django.py
#
# This file sets up and configures Django. It's used by scripts that need to
# execute as if running in a Django server.
import django
from django.conf import settings
from db import db_settings

def get_module_settings(module_name):
    module = globals().get(module_name, None)
    settings = {}
    if module:
        settings = {key: value for key, value in module.__dict__.items() \
            if not (key.startswith('__') or key.startswith('_')) and key.isupper()}
    return settings

def boot_django():
    if not settings.configured:
        overrides = get_module_settings('db_settings')
        settings.configure(**overrides)
        django.setup()


'''
Parse digital data from neural recording system into task data/messages/synchronization pulses
'''

import numpy as np

msgtype_mask = 0b0000111 << 8
auxdata_mask = 0b1111000 << 8
rawdata_mask = 0b11111111

MSG_TYPE_DATA = 0
MSG_TYPE_MESSAGE = 1
MSG_TYPE_REGISTER = 2
MSG_TYPE_REGISTER_SHAPE = 3
MSG_TYPE_ROW = 4
MSG_TYPE_ROWBYTE = 5

def parse_data(strobe_data):
    '''
    Parse out 'strobe' digital data into header/registrations + actual data
    '''
    reg_parsed_data = registrations(strobe_data)
    msg_parsed_data = messages(strobe_data)
    rowbyte_parsed_data = rowbyte(strobe_data)

    data = dict(messages=msg_parsed_data)
    for key in rowbyte_parsed_data:
        if key in reg_parsed_data:
            sys_name = reg_parsed_data[key][0]
            try:
                sys_dtype = np.dtype(eval(reg_parsed_data[key][1]))
            except:
                sys_dtype = reg_parsed_data[key][1]
            data[sys_name] = dict(row_ts=rowbyte_parsed_data[key])
        else:
            data[key] = dict(row_ts=rowbyte_parsed_data[key])

    return data

def _split(data, flip=False):
    '''
    Helper function to take the 16-bit integer saved in the neural data file 
    and map it back to the three fields of the message type (see docs on 
    communication protocol for details)

    Parameters
    ----------
    data : np.ndarray 
        Integer data and timestamps as stored in the neural data file when messages were sent during experiment

    Returns
    -------
    np.ndarray
        Raw message data split into the fields (type, aux, "payload")
    '''
    # If the data is a 1D array, extract the timestamps and the raw event codes
    if len(data.shape) < 2:
        data = np.array(data[data['chan'] == 257][['ts', 'unit']].tolist())
    msgs = data[:,1].astype(np.int16)
    
    if not flip:
        msgs = ~msgs # bit-flip the binary messages
    msgtype = np.right_shift(np.bitwise_and(msgs, msgtype_mask), 8).astype(np.uint8)
    auxdata = np.right_shift(np.bitwise_and(msgs, auxdata_mask), 8).astype(np.uint8)
    rawdata = np.bitwise_and(msgs, rawdata_mask)
    return np.vstack([data[:,0], msgtype, auxdata, rawdata]).T

def registrations(data, map_system=False):
    '''
    Parse the DIO data from the neural recording system to determine which 
    data sources were registered by the experiment software
    
    Parameters
    ---------
    data: np.array
        Digital bit data sent to the neural recording box. This data can either
        be a 1D record array with fields ('chan', 'unit', 'ts') or an N x 4
        regular array where the four columns are (timestamp, message type,
        auxiliary data, raw data)
    
    Returns
    -------
    systems : dict
        In the dictionary, keys are the ID # of the system (assigned sequentially
        during registration time when the task is initializing). Values are
        tuples of (name, dtype)
    '''
    if data.ndim < 2 or data.shape[1] != 4:
        if map_system:
            data = _split(data, flip=True)
        else:
            data = _split(data)

    ts, msgtype, auxdata, rawdata = data[:,0], data[:,1], data[:,2], data[:,3].astype(np.uint8)
    idx = msgtype == MSG_TYPE_REGISTER #data[:,1] == MSG_TYPE_REGISTER 
    shape_idx = msgtype == MSG_TYPE_REGISTER_SHAPE #data[:,1] == MSG_TYPE_REGISTER_SHAPE 

    regsysid = auxdata[idx] #data[idx][:,2] #should have more than
    #one value for more than one registration


    regshapeid = auxdata[shape_idx] #data[shape_idx][:,2]
    names = rawdata[idx] #data[idx][:,3].astype(np.uint8)
    
    dtype_data = rawdata[shape_idx]

    systems = dict()
    for sys in np.unique(regsysid):
        name = names[regsysid == sys].tostring()
        name = name[:-1] # Remove null terminator
        dtype = dtype_data[regshapeid == sys].tostring() #data[shape_idx][regshapeid == sys][:,3].astype(np.uint8).tostring()
        dtype = dtype[:-1] # Remove null terminator
        systems[sys] = name, dtype #name[:-1], dtype[:-1]
    #import pdb; pdb.set_trace()        
    return systems

def rowbyte(data, **kwargs):
    '''
    Parameters
    ----------
    data: np.array
        see docs for registrations for shape/dtype
    kwargs : dict
        see Docs for _split to see which kwargs are allowed

    Returns
    -------
    systems : dict
        In the dictionary, keys are the ID # of the system (assigned sequentially
        during registration time when the task is initializing). Values are
        np.array of shape (N, 2) where the columns are (timestamp, rawdata)
    '''
    if data.ndim < 2 or data.shape[1] != 4:
        data = _split(data, **kwargs)

    msgs = data[data[:,1] == MSG_TYPE_ROWBYTE]
    systems = dict()
    for i in np.unique(msgs[:,2]):
        systems[i] = msgs[msgs[:,2] == i][:,[0,-1]]
    return systems

def messages(data, **kwargs):
    '''
    Parse out any string messages sent byte-by-byte to the neural recording system

    Parameters
    ----------
    data : np.ndarray 
        Integer data and timestamps as stored in the neural data file when messages were sent during experiment
        OR, the result of the _split function

    Returns
    -------
    record array
        fields of 'time' and 'state' (message)
    '''
    if data.ndim < 2 or data.shape[1] != 4:
        data = _split(data, **kwargs)

    times = data[data[:,1] == 1, 0]
    names = data[data[:,1] == 1,-1].astype(np.uint8)

    tidx, = np.nonzero(names == 0)
    tidx = np.hstack([0, tidx+1])

    msgs = []
    for s, e in np.vstack([tidx[:-1], tidx[1:]-1]).T:
        msgs.append((times[s], names[s:e].tostring()))

    return np.array(msgs, dtype=[('time', np.float), ('state', 'S256')])




'''
HTML rendering 'view' functions for Django web interface. Retreive data from database to put into HTML format.
'''
import datetime
import pickle
import json
import traceback
import os

from django.template import RequestContext
from django.shortcuts import render
from django.http import HttpResponse

from . import exp_tracker

from config.rig_defaults import rig_settings

def main(request):
    return render(request, "main.html", dict(test_db=os.environ.get('BMI3D_TEST_DATABASE')))

def _list_exp_history(dbname='default', subject=None, task=None, max_entries=None, show_all_rigs=False, show_hidden=False):
    from . import models
    # from .models import TaskEntry, Task, Subject, Feature, Generator
    td = datetime.timedelta(days=60)

    filter_kwargs = {}
    if not show_hidden:
        filter_kwargs['visible'] = True
    if not show_all_rigs:
        filter_kwargs['rig_name'] = rig_settings['name']
    if not (subject is None) and isinstance(subject, str):
        filter_kwargs['subject__name'] = subject
    if not (task is None) and isinstance(task, str):
        filter_kwargs['task__name'] = task

    entries = models.TaskEntry.objects.using(dbname).filter(**filter_kwargs).order_by("-date")
    if isinstance(max_entries, int):
        entries = entries[:max_entries]

    for k in range(0, len(entries)):
        ent = entries[k]
        if k == 0 or not entries[k].date.date() == entries[k-1].date.date():
            ent.html_date = ent.date.date()
        else:
            ent.html_date = None
        ent.html_time = ent.date.time()

    ## Determine how many rows the date should span
    last = -1
    for k, ent in enumerate(entries[::-1]):
        if ent.html_date:
            ent.rowspan = k - last
            last = k

    filter_kwargs['template'] = True
    templates = models.TaskEntry.objects.using(dbname).filter(**filter_kwargs).order_by("-date")

    tasks = models.Task.objects.filter(visible=True).order_by("name")

    epoch = datetime.datetime.utcfromtimestamp(0)
    for entry in entries:
        tdiff = entry.date - epoch
        if tdiff.days % 2 == 0:
            entry.bgcolor = '#E1EEf4'
        else:
            entry.bgcolor = '#FFFFFF'

    features = models.Feature.objects.filter(visible=True).order_by("name")
    generators = models.Generator.objects.filter(visible=True).order_by("name")
    collections = models.TaskEntryCollection.objects.all().order_by("name")

    fields = dict(
        entries=entries,
        tasks=tasks,
        templates=templates,
        features=features,
        generators=generators,
        collections=collections,
        systems=models.System.objects.all(),
        n_entries=len(entries),
        show_hidden=show_hidden,
        n_hidden=len([e for e in entries if e.visible==False]),
        show_all_rigs=show_all_rigs,
        test_db=os.environ.get('BMI3D_TEST_DATABASE')
    )

    try:
        from config import bmiconfig
        fields['bmi_update_rates'] = bmiconfig.bmi_update_rates
        fields['state_spaces'] = bmiconfig.bmi_state_space_models
        fields['bmi_algorithms'] = bmiconfig.bmi_algorithms
        fields['extractors'] = bmiconfig.extractors
        fields['default_extractor'] = bmiconfig.default_extractor
        fields['pos_vars'] = bmiconfig.bmi_training_pos_vars      # 'pos_vars' indicates which column of the task HDF table to look at to extract kinematic data
        fields['kin_extractors'] = bmiconfig.kin_extractors       # post-processing methods on the selected kinematic variable
    except ImportError:
        pass
    return fields

def list_exp_history(request, **kwargs):
    '''
    Top-level view called when browser pointed at webroot

    Parameters
    ----------
    request: HTTPRequest instance
        No data needs to be extracted from this request

    Returns
    -------
    Django HTTPResponse instance
    '''
    from .models import TaskEntry, Task, Subject, Feature, Generator

    kwargs['show_hidden'] = request.GET.get('show_hidden') != None
    kwargs['show_all_rigs'] = request.GET.get('show_all_rigs') != None

    fields = _list_exp_history(**kwargs)
    fields['hostname'] = request.get_host()

    # this line is important--this is needed so the Track object knows if the task has ended in an error
    # TODO there's probably some better way of doing this within the multiprocessing lib (some code to run after the process has terminated)
    tracker = exp_tracker.get()
    tracker.update_alive()

    if tracker is not None:
        fields['status'] = tracker.get_status()
    if tracker.proc is not None and hasattr(tracker.proc, 'saveid'):
        fields['saveid'] = tracker.proc.saveid

    resp = render(request, 'list.html', fields)
    return resp

def setup_subjects(request):
    """view for experimenter to add new subjects"""
    from . import models
    from .models import TaskEntry, Task, Subject, Feature, Generator

    subjects = models.Subject.objects.all()
    experimenters = models.Experimenter.objects.all()

    return render(request, "setup_subjects.html", 
                  dict(subjects=subjects, 
                       experimenters=experimenters,
                       test_db=os.environ.get('BMI3D_TEST_DATABASE')))

def setup_tasks(request):
    """view for experimenter to add new tasks"""
    from . import models
    from .models import TaskEntry, Task, Subject, Feature, Generator

    tasks = models.Task.objects.all()

    return render(request, "setup_tasks.html", 
                  dict(tasks=tasks, test_db=os.environ.get('BMI3D_TEST_DATABASE')))

def setup_features(request):
    """view for experimenter to add new features"""
    from . import models
    from .models import TaskEntry, Task, Subject, Feature, Generator

    features = models.Feature.objects.all()

    # populate the list of built-in features which could be added
    from features import built_in_features
    built_in_feature_names = list(built_in_features.keys())

    return render(request, "setup_features.html",
        dict(active_features=features,
        active_feature_names=[feature.name for feature in features],
        built_in_feature_names=built_in_feature_names,
        test_db=os.environ.get('BMI3D_TEST_DATABASE')))        

def setup_parameters(request):
    """view for experimenter to add new global system parameters"""
    from . import models
    from .models import TaskEntry, Task, Subject, Feature, Generator

    systems = models.System.objects.all()

    # list of available databases
    from .. import db_settings
    databases = list(db_settings.DATABASES.keys())

    database_objs = []
    for db in databases:
        path = models.KeyValueStore.get('data_path', '--', dbname=db)
        database_objs.append({"name":db, "path":path})

    recording_sys = models.KeyValueStore.get("recording_sys")
    recording_sys_options = ['None', 'tdt', 'blackrock', 'plexon', 'ecube'] # TODO make this dynamic

    return render(request, "setup_parameters.html",
        dict(systems=systems, databases=database_objs, 
            recording_sys=recording_sys,
            recording_sys_options=recording_sys_options,
            test_db=os.environ.get('BMI3D_TEST_DATABASE')))

def setup(request):
    """Highest level "Setup" view"""
    return render(request, "setup_base.html", dict(test_db=os.environ.get('BMI3D_TEST_DATABASE')))

def _color_entries(entries):
    from .models import TaskEntry, Task, Subject, Feature, Generator

    epoch = datetime.datetime.utcfromtimestamp(0)

    last_tdiff = entries[0].date - epoch
    colors = ['#E1EEf4', '#FFFFFF']
    color_idx = 0
    for entry in entries:
        tdiff = entry.date - epoch
        if not (tdiff.days == last_tdiff.days):
            color_idx = (color_idx + 1) % 2
            last_tdiff = tdiff
        entry.bgcolor = colors[color_idx]

def get_sequence(request, idx):
    '''
    Pointing browser to WEBROOT/sequence_for/(?P<idx>\d+)/ returns a pickled
    file with the 'sequence' used in the specified id
    '''
    from .models import TaskEntry, Task, Subject, Feature, Generator
    entry = TaskEntry.objects.get(pk=idx)
    seq = pickle.loads(str(entry.sequence.sequence))
    log = json.loads(entry.report)
    num = len([l[2] for l in log if l[0] == "wait"])

    response = HttpResponse(pickle.dumps(seq[:num]), content_type='application/x-pickle')
    response['Content-Disposition'] = 'attachment; filename={subj}{time}_{idx}.pkl'.format(
        subj=entry.subject.name[:4].lower(),
        time="%04d%02d%02d"%(entry.date.year, entry.date.month, entry.date.day),
        idx=idx)
    return response

def link_data_files_view_generator(request, task_entry_id):
    from . import models
    from .models import TaskEntry, Task, Subject, Feature, Generator
    systems = models.System.objects.all()
    display_data = dict(systems=systems, task_entry_id=task_entry_id)
    return render(request, "link_data_files.html", display_data)

from django.views.decorators.csrf import csrf_exempt
@csrf_exempt
def link_data_files_response_handler(request, task_entry_id):
    from . import models
    file_path = request.POST["file_path"]
    raw_data = request.POST["raw_data"]
    raw_data_format = request.POST['raw_data_format']
    data_system_id = request.POST["data_system_id"]

    task_entry = models.TaskEntry.objects.get(id=task_entry_id)
    system = models.System.objects.get(id=data_system_id)
    if not os.path.isabs(file_path):
        sys_path = system.path
        if system.input_path is not None and system.input_path != "":
            sys_path = system.input_path
        else:
            sys_path = system.path
        file_path = os.path.join(sys_path, file_path)

    print("Adding file", file_path)

    if len(raw_data.strip()) > 0:
        # new file
        try:
            with open(file_path, 'w') as fp:
                fp.write(raw_data)
        except:
            import traceback
            traceback.print_exc()
            return HttpResponse("Unable to add new file")

    if os.path.exists(file_path):
        try:
            data_file = models.DataFile.create(system, task_entry, file_path, local=True, archived=False)
            return HttpResponse("Added new data file: {}".format(data_file.path))
        except:
            print("Error creating data file")
            traceback.print_exc()
            return HttpResponse("Error creating data file, see terminal for error message")
    else:
        return HttpResponse("File does not exist {}!".format(file_path))






